# Copyright 2015-2017 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import, division, print_function, unicode_literals

import datetime
import functools
import json
import os
import shutil
import tempfile
import time  # NOQA needed for some recordings

from unittest import TestCase

from botocore.exceptions import ClientError
from dateutil.tz import tzutc

from c7n.exceptions import PolicyValidationError
from c7n.executor import MainThreadExecutor
from c7n.resources import s3
from c7n.mu import LambdaManager
from c7n.ufuncs import s3crypt

from .common import (
    BaseTest,
    ConfigTest,
    event_data,
    skip_if_not_validating,
    functional,
    TestConfig as Config,
)


class RestoreCompletionTest(TestCase):

    def test_restore_complete(self):

        self.assertTrue(
            s3.restore_complete(
                (
                    'ongoing-request="false", '
                    'expiry-date="Fri, 23 Dec 2012 00:00:00 GMT"'
                )
            )
        )

        self.assertFalse(s3.restore_complete('ongoing-request="true"'))


class BucketScanLogTests(TestCase):

    def setUp(self):
        self.log_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.log_dir)
        self.log = s3.BucketScanLog(self.log_dir, "test")

    def test_scan_log(self):
        first_five = list(range(5))
        next_five = list(range(5, 10))
        with self.log:
            self.log.add(first_five)
            self.log.add(next_five)

        with open(self.log.path) as fh:
            data = json.load(fh)
            self.assertEqual(data, [first_five, next_five, []])


def destroyBucket(client, bucket):
    for o in client.list_objects(Bucket=bucket).get("Contents", []):
        client.delete_object(Bucket=bucket, Key=o["Key"])
    client.delete_bucket(Bucket=bucket)


def destroyVersionedBucket(client, bucket):
    for o in client.list_object_versions(Bucket=bucket).get("Versions", []):
        client.delete_object(Bucket=bucket, Key=o["Key"], VersionId=o["VersionId"])
    client.delete_bucket(Bucket=bucket)


def destroyBucketIfPresent(client, bucket):
    try:
        destroyVersionedBucket(client, bucket)
    except Exception as exc:
        response = getattr(
            exc, "response", {"ResponseMetadata": {"HTTPStatusCode": None}}
        )
        if response["ResponseMetadata"]["HTTPStatusCode"] != 404:
            raise


def generateBucketContents(s3, bucket, contents=None):
    default_contents = {
        "home.txt": "hello", "AWSLogs/2015/10/10": "out", "AWSLogs/2015/10/11": "spot"
    }
    if contents is None:
        contents = default_contents
    s3.Bucket(bucket)
    for k, v in contents.items():
        key = s3.Object(bucket, k)
        key.put(Body=v, ContentLength=len(v), ContentType="text/plain")


class BucketMetrics(BaseTest):

    def test_metrics(self):
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(s3, "S3_AUGMENT_TABLE", [])
        session_factory = self.replay_flight_data("test_s3_metrics")
        p = self.load_policy(
            {
                "name": "s3-obj-count",
                "resource": "s3",
                "filters": [
                    {
                        "type": "metrics",
                        "value": 10000,
                        "name": "NumberOfObjects",
                        "op": "greater-than",
                    }
                ],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 1)
        self.assertEqual(resources[0]["Name"], "custodian-skunk-trails")
        self.assertTrue("c7n.metrics" in resources[0])
        self.assertTrue("AWS/S3.NumberOfObjects.Average" in resources[0]["c7n.metrics"])


class BucketEncryption(BaseTest):

    def test_s3_bucket_encryption_filter(self):
        bname = "c7n-bucket-with-encryption"
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(s3, "S3_AUGMENT_TABLE", [])

        session_factory = self.replay_flight_data("test_s3_bucket_encryption_filter")

        client = session_factory().client("s3")
        client.create_bucket(Bucket=bname)
        self.addCleanup(client.delete_bucket, Bucket=bname)

        enc = {
            "Rules": [
                {"ApplyServerSideEncryptionByDefault": {"SSEAlgorithm": "AES256"}}
            ]
        }

        client.put_bucket_encryption(
            Bucket=bname, ServerSideEncryptionConfiguration=enc
        )

        p = self.load_policy(
            {
                "name": "s3-enc",
                "resource": "s3",
                "filters": [{"type": "bucket-encryption", "crypto": "AES256"}],
            },
            session_factory=session_factory,
        )
        resources = p.run() or []
        self.assertEqual(len(resources), 1)
        self.assertEqual(resources[0]["Name"], bname)

    def test_s3_bucket_encryption_filter_kms(self):
        bname = "c7n-bucket-with-encryption"
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(s3, "S3_AUGMENT_TABLE", [])

        session_factory = self.replay_flight_data(
            "test_s3_bucket_encryption_filter_kms"
        )

        client = session_factory().client("s3")
        client.create_bucket(Bucket=bname)
        self.addCleanup(client.delete_bucket, Bucket=bname)

        aws_alias = "arn:aws:kms:us-east-1:108891588060:key/079a6f7d-5f8a-4da1-a465-30aa099b9688"
        enc = {
            "Rules": [
                {
                    "ApplyServerSideEncryptionByDefault": {
                        "SSEAlgorithm": "aws:kms", "KMSMasterKeyID": aws_alias
                    }
                }
            ]
        }

        client.put_bucket_encryption(
            Bucket=bname, ServerSideEncryptionConfiguration=enc
        )

        p = self.load_policy(
            {
                "name": "s3-enc-kms",
                "resource": "s3",
                "filters": [
                    {
                        "type": "bucket-encryption",
                        "crypto": "aws:kms",
                        "key": "alias/aws/s3",
                    }
                ],
            },
            session_factory=session_factory,
        )
        resources = p.run() or []
        self.assertEqual(len(resources), 1)
        self.assertEqual(resources[0]["Name"], bname)

    def test_s3_filter_bucket_encryption_disabled(self):
        bname = "c7n-bucket-without-default-encryption"
        self.patch(s3.S3, "executor-factory", MainThreadExecutor)
        self.patch(s3, "S3_AUGMENT_TABLE", [])

        session_factory = self.replay_flight_data("test_s3_bucket_encryption_disabled")

        client = session_factory().client("s3")
        client.create_bucket(Bucket=bname)
        self.addCleanup(client.delete_bucket, Bucket=bname)

        p = self.load_policy(
            {
                "name": "s3-disabled-encryption",
                "resource": "s3",
                "filters": [
                    {"Name": bname}, {"type": "bucket-encryption", "state": False}
                ],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 1)
        self.assertRaises(ClientError, client.get_bucket_encryption, Bucket=bname)

        client.put_bucket_encryption(
            Bucket=bname,
            ServerSideEncryptionConfiguration={
                "Rules": [
                    {"ApplyServerSideEncryptionByDefault": {"SSEAlgorithm": "AES256"}}
                ]
            },
        )

        p = self.load_policy(
            {
                "name": "s3-disabled-encryption",
                "resource": "s3",
                "filters": [
                    {"Name": bname},
                    {"type": "bucket-encryption", "state": False, "crypto": "AES256"},
                ],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 0)


class BucketInventory(BaseTest):

    def test_s3_set_encrypted_inventory_sses3(self):
        bname = "custodian-inventory-test"
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(s3, "S3_AUGMENT_TABLE", [])

        session_factory = self.replay_flight_data(
            "test_s3_set_encrypted_inventory_sses3"
        )

        client = session_factory().client("s3")
        client.create_bucket(Bucket=bname)
        self.addCleanup(client.delete_bucket, Bucket=bname)

        p = self.load_policy(
            {
                "name": "s3-inv",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": [
                    {
                        "type": "set-inventory",
                        "destination": "inv-dest",
                        "name": "inv-name",
                        "state": "enabled",
                        "encryption": "SSES3",
                        "fields": ["Size", "EncryptionStatus"],
                    }
                ],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 1)

        invs = client.list_bucket_inventory_configurations(Bucket=bname).get(
            "InventoryConfigurationList"
        )
        self.assertTrue(invs)
        self.assertTrue(
            "SSES3" in invs[0]["Destination"]["S3BucketDestination"]["Encryption"]
        )
        self.assertTrue("EncryptionStatus" in invs[0]["OptionalFields"])

    def test_s3_set_encrypted_inventory_ssekms(self):
        bname = "custodian-inventory-test"
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(s3, "S3_AUGMENT_TABLE", [])

        session_factory = self.replay_flight_data(
            "test_s3_set_encrypted_inventory_ssekms"
        )

        client = session_factory().client("s3")
        client.create_bucket(Bucket=bname)
        self.addCleanup(client.delete_bucket, Bucket=bname)

        p = self.load_policy(
            {
                "name": "s3-inv",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": [
                    {
                        "type": "set-inventory",
                        "destination": "inv-dest",
                        "name": "inv-name",
                        "state": "enabled",
                        "encryption": "SSEKMS",
                        "key_id": "arn:valid:kms",
                        "fields": ["Size", "EncryptionStatus"],
                    }
                ],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 1)

        invs = client.list_bucket_inventory_configurations(Bucket=bname).get(
            "InventoryConfigurationList"
        )
        self.assertTrue(invs)
        self.assertTrue(
            "SSEKMS" in invs[0]["Destination"]["S3BucketDestination"]["Encryption"]
        )
        self.assertTrue("EncryptionStatus" in invs[0]["OptionalFields"])

    def test_s3_delete_inventory_inventory_not_set(self):
        bname = "delete_inventory"
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(s3, "S3_AUGMENT_TABLE", [])

        session_factory = self.replay_flight_data(
            "test_s3_delete_inventory_inventory_not_set"
        )

        client = session_factory().client("s3")
        client.create_bucket(Bucket=bname)
        self.addCleanup(client.delete_bucket, Bucket=bname)

        p = self.load_policy(
            {
                "name": "s3-inv",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": [
                    {
                        "type": "set-inventory",
                        "destination": "inv-dest",
                        "name": "inv-name",
                        "state": "absent",
                    }
                ],
            },
            session_factory=session_factory,
        )

        resources = p.run()
        self.assertEqual(len(resources), 1)
        inventoryConfigList = client.list_bucket_inventory_configurations(
            Bucket=bname
        ).get(
            "InventoryConfigurationList"
        )
        self.assertFalse(inventoryConfigList)

    @functional
    def test_inventory(self):
        bname = "custodian-test-data"
        inv_bname = "custodian-inv"
        inv_name = "something"

        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(s3, "S3_AUGMENT_TABLE", [])

        session_factory = self.replay_flight_data("test_s3_inventory")

        client = session_factory().client("s3")
        if self.recording:
            destroyBucketIfPresent(client, bname)
            destroyBucketIfPresent(client, inv_bname)
        client.create_bucket(
            Bucket=bname, CreateBucketConfiguration={"LocationConstraint": "us-east-2"}
        )
        client.create_bucket(
            Bucket=inv_bname,
            CreateBucketConfiguration={"LocationConstraint": "us-east-2"},
        )

        self.addCleanup(client.delete_bucket, Bucket=bname)
        self.addCleanup(client.delete_bucket, Bucket=inv_bname)

        inv = {
            "Destination": {
                "S3BucketDestination": {
                    "Bucket": "arn:aws:s3:::%s" % inv_bname,
                    "Format": "CSV",
                    "Prefix": "abcdef",
                }
            },
            "IsEnabled": True,
            "Id": inv_name,
            "IncludedObjectVersions": "All",
            "OptionalFields": ["LastModifiedDate"],
            "Schedule": {"Frequency": "Daily"},
        }

        client.put_bucket_inventory_configuration(
            Bucket=bname, Id=inv_name, InventoryConfiguration=inv
        )

        p = self.load_policy(
            {
                "name": "s3-inv",
                "resource": "s3",
                "filters": [{"Name": "custodian-test-data"}],
                "actions": [
                    {
                        "type": "set-inventory",
                        "destination": inv_bname,
                        "name": inv_name,
                    }
                ],
            },
            session_factory=session_factory,
        )
        self.assertEqual(len(p.run()), 1)
        invs = client.list_bucket_inventory_configurations(Bucket=bname).get(
            "InventoryConfigurationList"
        )
        self.assertTrue(invs)
        self.assertEqual(
            sorted(invs[0]["OptionalFields"]), ["LastModifiedDate", "Size"]
        )

        p = self.load_policy(
            {
                "name": "s3-inv",
                "resource": "s3",
                "filters": [{"Name": "custodian-test-data"}],
                "actions": [
                    {
                        "type": "set-inventory",
                        "destination": inv_bname,
                        "state": "absent",
                        "name": inv_name,
                    }
                ],
            },
            session_factory=session_factory,
        )

        self.assertEqual(len(p.run()), 1)
        self.assertFalse(
            client.list_bucket_inventory_configurations(Bucket=bname).get(
                "InventoryConfigurationList"
            )
        )


class BucketDelete(BaseTest):

    def test_delete_replicated_bucket(self):
        # the iam setup is a little for replication to duplicate in a test
        # preconditions - custodian-replicated and custodian-replicated-west
        # buckets setup with replication, we're deleting the custodian-replicated
        # bucket (source).
        bname = "custodian-replicated"
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(
            s3,
            "S3_AUGMENT_TABLE",
            [
                ("get_bucket_replication", "Replication", None, None),
                ("get_bucket_versioning", "Versioning", None, None),
            ],
        )
        session_factory = self.replay_flight_data("test_s3_delete_replicated_bucket")
        p = self.load_policy(
            {
                "name": "s3-delete-bucket",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": [{"type": "delete", "remove-contents": True}],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 1)

        session = session_factory()
        client = session.client("s3")
        buckets = set([b["Name"] for b in client.list_buckets()["Buckets"]])
        self.assertFalse(bname in buckets)

    @functional
    def test_delete_versioned_bucket(self):
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(
            s3,
            "S3_AUGMENT_TABLE",
            [("get_bucket_versioning", "Versioning", None, None)],
        )
        session_factory = self.replay_flight_data("test_s3_delete_versioned_bucket")
        session = session_factory()
        client = session.client("s3")
        s3_resource = session.resource("s3")
        bname = "custodian-byebye"
        if self.recording:
            destroyBucketIfPresent(client, bname)
        client.create_bucket(
            Bucket=bname, CreateBucketConfiguration={"LocationConstraint": "us-east-2"}
        )
        client.put_bucket_versioning(
            Bucket=bname, VersioningConfiguration={"Status": "Enabled"}
        )
        generateBucketContents(s3_resource, bname)
        # Generate some versions
        generateBucketContents(s3_resource, bname)

        upload_info = client.create_multipart_upload(Bucket=bname, Key="abcdef12345")
        client.upload_part(
            Body="1" * 1024,
            Bucket=bname,
            Key="abcdef12345",
            PartNumber=1,
            UploadId=upload_info["UploadId"],
        )

        p = self.load_policy(
            {
                "name": "s3-delete-bucket",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": [{"type": "delete", "remove-contents": True}],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        if self.recording:
            time.sleep(60)
        self.assertEqual(len(resources), 1)
        buckets = set([b["Name"] for b in client.list_buckets()["Buckets"]])
        self.assertFalse(bname in buckets)

    @functional
    def test_delete_bucket(self):
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(s3.DeleteBucket, "executor_factory", MainThreadExecutor)
        self.patch(s3, "S3_AUGMENT_TABLE", [])
        session_factory = self.replay_flight_data("test_s3_delete_bucket")
        session = session_factory()
        client = session.client("s3")
        bname = "custodian-byebye"

        if self.recording:
            destroyBucketIfPresent(client, bname)
        client.create_bucket(
            Bucket=bname, CreateBucketConfiguration={"LocationConstraint": "us-east-2"}
        )
        generateBucketContents(session.resource("s3"), bname)

        p = self.load_policy(
            {
                "name": "s3-delete-bucket",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": [{"type": "delete", "remove-contents": True}],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 1)
        buckets = set([b["Name"] for b in client.list_buckets()["Buckets"]])
        self.assertFalse(bname in buckets)

    def test_delete_bucket_with_failure(self):
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(s3.DeleteBucket, "executor_factory", MainThreadExecutor)
        self.patch(s3, "S3_AUGMENT_TABLE", [])

        session_factory = self.replay_flight_data("test_s3_delete_bucket_with_failure")
        session = session_factory()
        client = session.client("s3")
        bname = "custodian-perm-denied"
        client.create_bucket(Bucket=bname)
        generateBucketContents(session.resource("s3"), bname)

        # This bucket policy prevents viewing contents
        policy = {
            "Version": "2012-10-17",
            "Id": "Policy1487359365244",
            "Statement": [
                {
                    "Sid": "Stmt1487359361981",
                    "Effect": "Deny",
                    "Principal": "*",
                    "Action": "s3:DeleteBucket",
                    "Resource": "arn:aws:s3:::{}".format(bname),
                }
            ],
        }
        client.put_bucket_policy(Bucket=bname, Policy=json.dumps(policy))

        p = self.load_policy(
            {
                "name": "s3-delete-bucket",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": [{"type": "delete", "remove-contents": True}],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 1)
        buckets = set([b["Name"] for b in client.list_buckets()["Buckets"]])
        self.assertIn(bname, buckets)

        # Make sure file got written
        denied_file = os.path.join(p.ctx.log_dir, "denied.json")
        self.assertIn(bname, open(denied_file).read())
        #
        # Now delete it for real
        #
        client.delete_bucket_policy(Bucket=bname)
        resources = p.run()
        self.assertEqual(len(resources), 1)
        buckets = set([b["Name"] for b in client.list_buckets()["Buckets"]])
        self.assertFalse(bname in buckets)


class BucketTag(BaseTest):

    @functional
    def test_tag_bucket(self):
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(s3.EncryptExtantKeys, "executor_factory", MainThreadExecutor)
        self.patch(
            s3, "S3_AUGMENT_TABLE", [("get_bucket_tagging", "Tags", [], "TagSet")]
        )
        session_factory = self.replay_flight_data("test_s3_tag")
        session = session_factory()
        client = session.client("s3")
        bname = "custodian-tagger"
        if self.recording:
            destroyBucketIfPresent(client, bname)
        client.create_bucket(
            Bucket=bname, CreateBucketConfiguration={"LocationConstraint": "us-east-2"}
        )
        self.addCleanup(destroyBucket, client, bname)
        client.put_bucket_tagging(
            Bucket=bname,
            Tagging={
                "TagSet": [
                    {"Key": "rudolph", "Value": "reindeer"},
                    {"Key": "platform", "Value": "lxwee"},
                ]
            },
        )

        p = self.load_policy(
            {
                "name": "s3-tagger",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": [
                    {
                        "type": "tag",
                        "tags": {"borrowed": "new", "platform": "serverless"},
                    }
                ],
            },
            session_factory=session_factory,
        )

        resources = p.run()
        self.assertEqual(len(resources), 1)
        tags = {
            t["Key"]: t["Value"]
            for t in client.get_bucket_tagging(Bucket=bname)["TagSet"]
        }
        self.assertEqual(
            {"rudolph": "reindeer", "platform": "serverless", "borrowed": "new"}, tags
        )


class S3ConfigSource(ConfigTest):

    maxDiff = None

    @functional
    def test_normalize(self):
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        augments = list(s3.S3_AUGMENT_TABLE)
        augments.remove(("get_bucket_location", "Location", None, None))
        self.patch(s3, "S3_AUGMENT_TABLE", augments)

        bname = "custodian-test-data-23"
        session_factory = self.replay_flight_data("test_s3_normalize")
        session = session_factory()

        queue_url = self.initialize_config_subscriber(session)
        client = session.client("s3")
        if self.recording:
            destroyBucketIfPresent(client, bname)
        client.create_bucket(
            Bucket=bname, CreateBucketConfiguration={"LocationConstraint": "us-east-2"}
        )
        self.addCleanup(destroyBucket, client, bname)

        sns = session.client("sns")
        notify_topic = sns.create_topic(Name=bname).get("TopicArn")
        sns.set_topic_attributes(
            TopicArn=notify_topic,
            AttributeName="Policy",
            AttributeValue=json.dumps(
                {
                    "Statement": [
                        {
                            "Action": "SNS:Publish",
                            "Effect": "Allow",
                            "Resource": notify_topic,
                            "Principal": {"Service": "s3.amazonaws.com"},
                        }
                    ]
                }
            ),
        )
        self.addCleanup(sns.delete_topic, TopicArn=notify_topic)

        public = "http://acs.amazonaws.com/groups/global/AuthenticatedUsers"
        client.put_bucket_acl(
            Bucket=bname,
            AccessControlPolicy={
                "Owner": {
                    "DisplayName": "mandeep.bal",
                    "ID": "e7c8bb65a5fc49cf906715eae09de9e4bb7861a96361ba79b833aa45f6833b15",
                },
                "Grants": [
                    {"Grantee": {"Type": "Group", "URI": public}, "Permission": "READ"},
                    {
                        "Grantee": {
                            "Type": "Group",
                            "URI": "http://acs.amazonaws.com/groups/s3/LogDelivery",
                        },
                        "Permission": "WRITE",
                    },
                    {
                        "Grantee": {
                            "Type": "Group",
                            "URI": "http://acs.amazonaws.com/groups/s3/LogDelivery",
                        },
                        "Permission": "READ_ACP",
                    },
                ],
            },
        )
        client.put_bucket_tagging(
            Bucket=bname,
            Tagging={
                "TagSet": [
                    {"Key": "rudolph", "Value": "rabbit"},
                    {"Key": "platform", "Value": "tyre"},
                ]
            },
        )
        client.put_bucket_logging(
            Bucket=bname,
            BucketLoggingStatus={
                "LoggingEnabled": {"TargetBucket": bname, "TargetPrefix": "s3-logs/"}
            },
        )
        client.put_bucket_versioning(
            Bucket=bname, VersioningConfiguration={"Status": "Enabled"}
        )
        client.put_bucket_accelerate_configuration(
            Bucket=bname, AccelerateConfiguration={"Status": "Enabled"}
        )
        client.put_bucket_website(
            Bucket=bname,
            WebsiteConfiguration={"IndexDocument": {"Suffix": "index.html"}},
        )
        client.put_bucket_policy(
            Bucket=bname,
            Policy=json.dumps(
                {
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Sid": "Zebra",
                            "Effect": "Deny",
                            "Principal": "*",
                            "Action": "s3:PutObject",
                            "Resource": "arn:aws:s3:::%s/*" % bname,
                            "Condition": {
                                "StringNotEquals": {
                                    "s3:x-amz-server-side-encryption": [
                                        "AES256", "aws:kms"
                                    ]
                                }
                            },
                        }
                    ],
                }
            ),
        )
        client.put_bucket_notification_configuration(
            Bucket=bname,
            NotificationConfiguration={
                "TopicConfigurations": [
                    {
                        "Id": bname,
                        "TopicArn": notify_topic,
                        "Events": ["s3:ObjectCreated:*"],
                        "Filter": {
                            "Key": {
                                "FilterRules": [{"Name": "prefix", "Value": "s3-logs/"}]
                            }
                        },
                    }
                ]
            },
        )

        p = self.load_policy(
            {"name": "s3-inv", "resource": "s3", "filters": [{"Name": bname}]},
            session_factory=session_factory,
        )

        manager = p.load_resource_manager()
        resource_a = manager.get_resources([bname])[0]
        results = self.wait_for_config(session, queue_url, bname)
        resource_b = s3.ConfigS3(manager).load_resource(results[0])
        self.maxDiff = None

        for k in ("Logging", "Policy", "Versioning", "Name", "Website"):
            self.assertEqual(resource_a[k], resource_b[k])

        self.assertEqual(
            {t["Key"]: t["Value"] for t in resource_a.get("Tags")},
            {t["Key"]: t["Value"] for t in resource_b.get("Tags")},
        )

    def test_config_normalize_notification(self):
        event = event_data("s3-rep-and-notify.json", "config")
        p = self.load_policy({"name": "s3cfg", "resource": "s3"})
        source = p.resource_manager.get_source("config")
        resource = source.load_resource(event)
        self.assertEqual(
            resource["Notification"],
            {
                u"TopicConfigurations": [
                    {
                        u"Filter": {
                            u"Key": {
                                u"FilterRules": [{u"Name": "Prefix", u"Value": "oids/"}]
                            }
                        },
                        u"Id": "rabbit",
                        u"TopicArn": "arn:aws:sns:us-east-1:644160558196:custodian-test-data-22",
                        u"Events": [
                            "s3:ReducedRedundancyLostObject",
                            "s3:ObjectCreated:CompleteMultipartUpload",
                        ],
                    }
                ],
                u"LambdaFunctionConfigurations": [
                    {
                        u"Filter": {
                            u"Key": {
                                u"FilterRules": [{u"Name": "Prefix", u"Value": "void/"}]
                            }
                        },
                        u"LambdaFunctionArn": (
                            "arn:aws:lambda:us-east-1:644160558196:function:lambdaenv"
                        ),
                        u"Id": "ZDAzZDViMTUtNGU3MS00ZWIwLWI0MzgtOTZiMWQ3ZWNkZDY1",
                        u"Events": ["s3:ObjectRemoved:Delete"],
                    }
                ],
                u"QueueConfigurations": [
                    {
                        u"Filter": {
                            u"Key": {
                                u"FilterRules": [
                                    {u"Name": "Prefix", u"Value": "images/"}
                                ]
                            }
                        },
                        u"Id": "OGQ5OTAyNjYtYjBmNy00ZTkwLWFiMjUtZjE4ODBmYTgwNTE0",
                        u"QueueArn": "arn:aws:sqs:us-east-1:644160558196:test-queue",
                        u"Events": ["s3:ObjectCreated:*"],
                    }
                ],
            },
        )

    def test_config_normalize_lifecycle_null_predicate(self):
        event = event_data("s3-lifecycle-null-predicate.json", "config")
        p = self.load_policy({"name": "s3cfg", "resource": "s3"})
        source = p.resource_manager.get_source("config")
        resource = source.load_resource(event)
        rule = resource["Lifecycle"]["Rules"][0]
        self.assertEqual(
            rule,
            {
                "AbortIncompleteMultipartUpload": {"DaysAfterInitiation": 1},
                "Expiration": {"Days": 1},
                "ID": "RemoveAbortedUploads",
                "NoncurrentVersionExpiration": {"NoncurrentDays": -1},
                "Status": "Enabled",
            },
        )

    def test_config_normalize_lifecycle_and_predicate(self):
        event = event_data("s3-lifecycle-and-predicate.json", "config")
        p = self.load_policy({"name": "s3cfg", "resource": "s3"})
        source = p.resource_manager.get_source("config")
        resource = source.load_resource(event)
        rfilter = resource["Lifecycle"]["Rules"][0]["Filter"]

        self.assertEqual(rfilter["And"]["Prefix"], "docs/")
        self.assertEqual(
            rfilter["And"]["Tags"],
            [
                {"Value": "Archive", "Key": "Workflow"},
                {"Value": "Complete", "Key": "State"},
            ],
        )

    def test_config_normalize_lifecycle(self):
        event = event_data("s3-lifecycle.json", "config")
        p = self.load_policy({"name": "s3cfg", "resource": "s3"})
        source = p.resource_manager.get_source("config")
        resource = source.load_resource(event)
        self.assertEqual(
            resource["Lifecycle"],
            {
                "Rules": [
                    {
                        "Status": "Enabled",
                        "NoncurrentVersionExpiration": {"NoncurrentDays": 545},
                        "Filter": {"Prefix": "docs/"},
                        "Transitions": [{"Days": 30, "StorageClass": "STANDARD_IA"}],
                        "Expiration": {"ExpiredObjectDeleteMarker": True},
                        "AbortIncompleteMultipartUpload": {"DaysAfterInitiation": 7},
                        "NoncurrentVersionTransitions": [
                            {"NoncurrentDays": 180, "StorageClass": "GLACIER"}
                        ],
                        "ID": "Docs",
                    }
                ]
            },
        )

    def test_config_normalize_replication(self):
        event = event_data("s3-rep-and-notify.json", "config")
        p = self.load_policy({"name": "s3cfg", "resource": "s3"})
        source = p.resource_manager.get_source("config")
        resource = source.load_resource(event)
        self.assertEqual(
            resource["Replication"],
            {
                u"ReplicationConfiguration": {
                    u"Rules": [
                        {
                            u"Status": "Enabled",
                            u"Prefix": "",
                            u"Destination": {u"Bucket": "arn:aws:s3:::testing-west"},
                            u"ID": "testing-99",
                        }
                    ],
                    u"Role": (
                        "arn:aws:iam::644160558196:role"
                        "/custodian-replicated-custodian-replicated"
                        "-west-s3-repl-role"
                    ),
                }
            },
        )

    def test_config_normalize_website_redirect(self):
        event = event_data("s3-website-redirect.json", "config")
        p = self.load_policy({"name": "s3cfg", "resource": "s3"})
        source = p.resource_manager.get_source("config")
        self.maxDiff = None
        resource = source.load_resource(event)
        self.assertEqual(
            resource["Website"],
            {
                "RedirectAllRequestsTo": {
                    "HostName": "www.google.com/", "Protocol": "https"
                }
            },
        )

    def test_config_normalize_website(self):
        event = event_data("s3-website.json", "config")
        p = self.load_policy({"name": "s3cfg", "resource": "s3"})
        source = p.resource_manager.get_source("config")
        self.maxDiff = None
        resource = source.load_resource(event)
        self.assertEqual(
            resource["Website"],
            {
                u"IndexDocument": {u"Suffix": "index.html"},
                u"RoutingRules": [
                    {
                        u"Redirect": {u"ReplaceKeyWith": "error.html"},
                        u"Condition": {
                            u"HttpErrorCodeReturnedEquals": "404",
                            u"KeyPrefixEquals": "docs/",
                        },
                    }
                ],
            },
        )

    def test_load_item_resource(self):
        event = event_data("s3.json", "config")
        p = self.load_policy({"name": "s3cfg", "resource": "s3"})
        source = p.resource_manager.get_source("config")
        self.maxDiff = None
        resource = source.load_resource(event)
        resource.pop("CreationDate")
        self.assertEqual(
            {"Planet": "Earth", "Verbose": "Game"},
            {t["Key"]: t["Value"] for t in resource.pop("Tags")},
        )
        self.assertEqual(
            resource,
            {
                "Location": {"LocationConstraint": u"us-east-2"},
                "Name": u"config-rule-sanity",
                "Lifecycle": None,
                "Website": None,
                "Policy": None,
                "Replication": None,
                "Versioning": None,
                "Logging": None,
                "Notification": None,
                "Acl": {
                    "Owner": {
                        "ID": u"e7c8bb65a5fc49cf906715eae09de9e4bb7861a96361ba79b833aa45f6833b15"
                    },
                    "Grants": [
                        {
                            "Grantee": {
                                "Type": "CanonicalUser",
                                "ID": (
                                    "e7c8bb65a5fc49cf906715eae09de"
                                    "9e4bb7861a96361ba79b833aa45f6833b15"
                                ),
                            },
                            "Permission": "FULL_CONTROL",
                        }
                    ],
                },
            },
        )

    def test_load_item_resource_config_event(self):
        event = event_data("s3-from-rule.json", "config")
        p = self.load_policy({"name": "s3cfg", "resource": "s3"})
        source = p.resource_manager.get_source("config")
        self.maxDiff = None

        resource_config = json.loads(event["invokingEvent"])["configurationItem"]
        resource = source.load_resource(resource_config)
        self.assertEqual(
            resource,
            {
                u"Acl": {
                    u"Grants": [
                        {
                            u"Grantee": {
                                u"ID": (
                                    "e7c8bb65a5fc49cf906715eae09de9e4"
                                    "bb7861a96361ba79b833aa45f6833b15"
                                ),
                                u"Type": u"CanonicalUser",
                            },
                            u"Permission": u"FULL_CONTROL",
                        }
                    ],
                    u"Owner": {
                        u"DisplayName": u"mandeep.bal",
                        u"ID": u"e7c8bb65a5fc49cf906715eae09de9e4bb7861a96361ba79b833aa45f6833b15",
                    },
                },
                u"CreationDate": datetime.datetime(
                    2017, 9, 15, 2, 5, 40, tzinfo=tzutc()
                ),
                u"Lifecycle": None,
                u"Location": None,
                u"Logging": None,
                u"Name": u"c7n-fire-logs",
                u"Notification": {},
                u"Policy": None,
                u"Replication": None,
                u"Tags": [],
                u"Versioning": None,
                u"Website": None,
            },
        )


class BucketPolicyStatements(BaseTest):

    @functional
    def test_policy(self):
        bname = "custodian-test-data"
        sid = "CustodianTest"

        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(
            s3, "S3_AUGMENT_TABLE", [("get_bucket_policy", "Policy", None, "Policy")]
        )

        session_factory = self.replay_flight_data("test_s3_policy_statements")

        client = session_factory().client("s3")
        if self.recording:
            destroyBucketIfPresent(client, bname)
        client.create_bucket(
            Bucket=bname, CreateBucketConfiguration={"LocationConstraint": "us-east-2"}
        )

        self.addCleanup(client.delete_bucket, Bucket=bname)

        p = self.load_policy(
            {
                "name": "s3-policy-statements",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": [
                    {
                        "type": "set-statements",
                        "statements": [
                            {
                                "Sid": sid,
                                "Effect": "Deny",
                                "Action": "s3:GetObject",
                                "Principal": {"AWS": "*"},
                                "Resource": "arn:aws:s3:::{bucket_name}/*",
                                "Condition": {"Bool": {"aws:SecureTransport": False}},
                            }
                        ],
                    }
                ],
            },
            session_factory=session_factory,
        )
        self.assertEqual(len(p.run()), 1)

        policy = client.get_bucket_policy(Bucket=bname).get("Policy")
        policy = json.loads(policy)

        self.assertTrue(len(policy["Statement"]) > 0)
        self.assertTrue(
            len([s for s in policy["Statement"] if s["Sid"] == sid and
              s["Resource"] == "arn:aws:s3:::%s/*" % (bname)]) == 1
        )

    @functional
    def test_policy_no_change(self):
        bname = "custodian-test-data"

        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(
            s3, "S3_AUGMENT_TABLE", [("get_bucket_policy", "Policy", None, "Policy")]
        )

        session_factory = self.replay_flight_data("test_s3_policy_statements_no_change")

        client = session_factory().client("s3")
        if self.recording:
            destroyBucketIfPresent(client, bname)
        client.create_bucket(
            Bucket=bname, CreateBucketConfiguration={"LocationConstraint": "us-east-2"}
        )

        self.addCleanup(client.delete_bucket, Bucket=bname)

        p = self.load_policy(
            {
                "name": "s3-policy-statements",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": [{"type": "set-statements", "statements": []}],
            },
            session_factory=session_factory,
        )
        self.assertEqual(len(p.run()), 1)


class S3Test(BaseTest):

    def test_multipart_large_file(self):
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(s3.EncryptExtantKeys, "executor_factory", MainThreadExecutor)
        self.patch(s3, "S3_AUGMENT_TABLE", [])
        self.patch(s3, "MAX_COPY_SIZE", (1024 * 1024 * 6.1))
        session_factory = self.replay_flight_data("test_s3_multipart_file")
        session = session_factory()
        client = session.client("s3")
        bname = "custodian-largef-test"
        key = "hello"
        client.create_bucket(Bucket=bname)
        self.addCleanup(destroyBucket, client, bname)

        class wrapper(object):

            def __init__(self, d, length):
                self.d = d
                self.len = length
                self.counter = length

            def read(self, size):
                if self.counter == 0:
                    return ""
                if size > self.counter:
                    size = self.counter
                    self.counter = 0
                else:
                    self.counter -= size
                return self.d.read(size)

            def seek(self, offset, whence=0):
                if whence == 2 and offset == 0:
                    self.counter = 0
                elif whence == 0 and offset == 0:
                    self.counter = self.len

            def tell(self):
                return self.len - self.counter

        size = 1024 * 1024 * 16
        client.put_object(
            Bucket=bname,
            Key=key,
            Metadata={"planet": "earth"},
            Body=wrapper(open("/dev/zero"), size),
            ContentLength=size,
        )
        info = client.head_object(Bucket=bname, Key=key)
        p = self.load_policy(
            {
                "name": "encrypt-obj",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": ["encrypt-keys"],
            },
            session_factory=session_factory,
        )
        p.run()
        post_info = client.head_object(Bucket=bname, Key="hello")
        self.assertTrue("ServerSideEncryption" in post_info)
        self.assertEqual(post_info["Metadata"], {"planet": "earth"})
        # etags on multipart do not reflect md5 :-(
        self.assertTrue(info["ContentLength"], post_info["ContentLength"])

    def test_self_log(self):
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(
            s3,
            "S3_AUGMENT_TABLE",
            [("get_bucket_logging", "Logging", None, "LoggingEnabled")],
        )
        session_factory = self.replay_flight_data("test_s3_self_log_target")
        session = session_factory()
        client = session.client("s3")
        bname = "custodian-log-test"
        client.create_bucket(Bucket=bname)
        self.addCleanup(client.delete_bucket, Bucket=bname)
        client.put_bucket_acl(
            Bucket=bname,
            AccessControlPolicy={
                "Owner": {
                    "DisplayName": "k_vertigo",
                    "ID": "904fc4c4790937100e9eb293a15e6a0a1f265a064888055b43d030034f8881ee",
                },
                "Grants": [
                    {
                        "Grantee": {
                            "Type": "Group",
                            "URI": "http://acs.amazonaws.com/groups/s3/LogDelivery",
                        },
                        "Permission": "WRITE",
                    },
                    {
                        "Grantee": {
                            "Type": "Group",
                            "URI": "http://acs.amazonaws.com/groups/s3/LogDelivery",
                        },
                        "Permission": "READ_ACP",
                    },
                ],
            },
        )
        client.put_bucket_logging(
            Bucket=bname,
            BucketLoggingStatus={
                "LoggingEnabled": {"TargetBucket": bname, "TargetPrefix": "s3-logs/"}
            },
        )
        p = self.load_policy(
            {
                "name": "s3-log-targets",
                "resource": "s3",
                "filters": [{"Name": bname}, {"type": "is-log-target", "self": True}],
            },
            session_factory=session_factory,
        )

        resources = p.run()
        names = [b["Name"] for b in resources]
        self.assertEqual(names[0], bname)
        self.assertEqual(len(names), 1)

    def test_log_target(self):
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(
            s3,
            "S3_AUGMENT_TABLE",
            [("get_bucket_logging", "Logging", None, "LoggingEnabled")],
        )
        session_factory = self.replay_flight_data("test_s3_log_target")
        session = session_factory()
        client = session.client("s3")
        bname = "custodian-log-test"
        client.create_bucket(Bucket="custodian-log-test")
        self.addCleanup(client.delete_bucket, Bucket=bname)
        client.put_bucket_acl(
            Bucket=bname,
            AccessControlPolicy={
                "Owner": {
                    "DisplayName": "k_vertigo",
                    "ID": "904fc4c4790937100e9eb293a15e6a0a1f265a064888055b43d030034f8881ee",
                },
                "Grants": [
                    {
                        "Grantee": {
                            "Type": "Group",
                            "URI": "http://acs.amazonaws.com/groups/s3/LogDelivery",
                        },
                        "Permission": "WRITE",
                    },
                    {
                        "Grantee": {
                            "Type": "Group",
                            "URI": "http://acs.amazonaws.com/groups/s3/LogDelivery",
                        },
                        "Permission": "READ_ACP",
                    },
                ],
            },
        )
        client.put_bucket_logging(
            Bucket=bname,
            BucketLoggingStatus={
                "LoggingEnabled": {"TargetBucket": bname, "TargetPrefix": "s3-logs/"}
            },
        )
        p = self.load_policy(
            {"name": "s3-log-targets", "resource": "s3", "filters": ["is-log-target"]},
            session_factory=session_factory,
        )
        resources = p.run()
        names = [b["Name"] for b in resources]
        self.assertTrue(bname in names)

    def test_has_statement(self):
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(
            s3.MissingPolicyStatementFilter, "executor_factory", MainThreadExecutor
        )
        self.patch(
            s3, "S3_AUGMENT_TABLE", [("get_bucket_policy", "Policy", None, "Policy")]
        )
        session_factory = self.replay_flight_data("test_s3_has_statement")
        bname = "custodian-policy-test"
        session = session_factory()
        client = session.client("s3")
        client.create_bucket(Bucket=bname)
        self.addCleanup(destroyBucket, client, bname)
        client.put_bucket_policy(
            Bucket=bname,
            Policy=json.dumps(
                {
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Sid": "Zebra",
                            "Effect": "Deny",
                            "Principal": "*",
                            "Action": "s3:PutObject",
                            "Resource": "arn:aws:s3:::%s/*" % bname,
                            "Condition": {
                                "StringNotEquals": {
                                    "s3:x-amz-server-side-encryption": [
                                        "AES256", "aws:kms"
                                    ]
                                }
                            },
                        },
                        {
                            "Sid": "Zebra2",
                            "Effect": "Deny",
                            "Principal": "arn:aws:iam::644160558196:root",
                            "Action": "s3:PutObject",
                            "Resource": "arn:aws:s3:::%s/*" % bname,
                        },
                    ],
                }
            ),
        )
        p = self.load_policy(
            {
                "name": "s3-has-policy",
                "resource": "s3",
                "filters": [
                    {"Name": bname},
                    {"type": "has-statement", "statement_ids": ["Zebra"]},
                ],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 1)

    def test_has_statement_policy(self):
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(
            s3.MissingPolicyStatementFilter, "executor_factory", MainThreadExecutor
        )
        self.patch(
            s3, "S3_AUGMENT_TABLE", [("get_bucket_policy", "Policy", None, "Policy")]
        )
        session_factory = self.replay_flight_data("test_s3_has_statement")
        bname = "custodian-policy-test"
        p = self.load_policy(
            {
                "name": "s3-has-policy",
                "resource": "s3",
                "filters": [
                    {"Name": bname},
                    {
                        "type": "has-statement",
                        "statements": [
                            {
                                "Effect": "Deny",
                                "Action": "s3:PutObject",
                                "Principal": "*",
                            }
                        ],
                    },
                ],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 1)

    def test_has_statement_similar_policies(self):
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(
            s3.MissingPolicyStatementFilter, "executor_factory", MainThreadExecutor
        )
        self.patch(
            s3, "S3_AUGMENT_TABLE", [("get_bucket_policy", "Policy", None, "Policy")]
        )
        session_factory = self.replay_flight_data("test_s3_has_statement")
        bname = "custodian-policy-test"

        p = self.load_policy(
            {
                "name": "s3-has-policy",
                "resource": "s3",
                "filters": [
                    {"Name": bname},
                    {
                        "type": "has-statement",
                        "statements": [{"Effect": "Deny", "Action": "s3:PutObject"}],
                    },
                ],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 1)

    def test_no_encryption_statement(self):
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(
            s3.MissingPolicyStatementFilter, "executor_factory", MainThreadExecutor
        )
        self.patch(
            s3, "S3_AUGMENT_TABLE", [("get_bucket_policy", "Policy", None, "Policy")]
        )
        session_factory = self.replay_flight_data("test_s3_no_encryption_statement")
        bname = "custodian-encryption-test"
        session = session_factory()
        client = session.client("s3")
        client.create_bucket(Bucket=bname)
        self.addCleanup(destroyBucket, client, bname)
        client.put_bucket_policy(
            Bucket=bname,
            Policy=json.dumps(
                {
                    "Version": "2017-3-28",
                    "Statement": [
                        {
                            "Sid": "RequiredEncryptedObject",
                            "Effect": "Allow",
                            "Principal": "*",
                            "Action": "s3:PutObject",
                            "Resource": "arn:aws:s3:::%s/*" % bname,
                            "Condition": {
                                "StringNotEquals": {
                                    "s3:x-amz-server-side-encryption": [
                                        "AES256", "aws:kms"
                                    ]
                                }
                            },
                        }
                    ],
                }
            ),
        )
        p = self.load_policy(
            {
                "name": "s3-no-encryption-policy",
                "resource": "s3",
                "filters": [{"Name": bname}, {"type": "no-encryption-statement"}],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 1)

    def test_missing_policy_statement(self):
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(
            s3.MissingPolicyStatementFilter, "executor_factory", MainThreadExecutor
        )
        self.patch(
            s3, "S3_AUGMENT_TABLE", [("get_bucket_policy", "Policy", None, "Policy")]
        )
        session_factory = self.replay_flight_data("test_s3_missing_policy")
        bname = "custodian-encrypt-test"
        session = session_factory()
        client = session.client("s3")
        client.create_bucket(Bucket=bname)
        self.addCleanup(destroyBucket, client, bname)
        client.put_bucket_policy(
            Bucket=bname,
            Policy=json.dumps(
                {
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Sid": "Zebra",
                            "Effect": "Deny",
                            "Principal": "*",
                            "Action": "s3:PutObject",
                            "Resource": "arn:aws:s3:::%s/*" % bname,
                            "Condition": {
                                "StringNotEquals": {
                                    "s3:x-amz-server-side-encryption": [
                                        "AES256", "aws:kms"
                                    ]
                                }
                            },
                        }
                    ],
                }
            ),
        )
        p = self.load_policy(
            {
                "name": "encrypt-keys",
                "resource": "s3",
                "filters": [
                    {"Name": bname},
                    {
                        "type": "missing-policy-statement",
                        "statement_ids": ["RequireEncryptedPutObject"],
                    },
                ],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 1)

    def test_enable_versioning(self):
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(
            s3,
            "S3_AUGMENT_TABLE",
            [("get_bucket_versioning", "Versioning", None, None)],
        )
        session_factory = self.replay_flight_data("test_s3_enable_versioning")
        bname = "superduper-and-magic"
        session = session_factory()
        client = session.client("s3")
        client.create_bucket(Bucket=bname)
        self.addCleanup(destroyBucket, client, bname)
        p = self.load_policy(
            {
                "name": "s3-version",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": ["toggle-versioning"],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 1)
        self.assertEqual(resources[0]["Name"], bname)

        # eventual consistency fun for recording
        # time.sleep(10)
        versioning = client.get_bucket_versioning(Bucket=bname)["Status"]
        self.assertEqual("Enabled", versioning)

        # running against a bucket with versioning already on
        # is idempotent
        resources = p.run()
        self.assertEqual(len(resources), 1)

        p = self.load_policy(
            {
                "name": "s3-version",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": [{"type": "toggle-versioning", "enabled": False}],
            },
            session_factory=session_factory,
        )

        resources = p.run()
        self.assertEqual(len(resources), 1)

        # eventual consistency fun for recording
        # time.sleep(10)
        versioning = client.get_bucket_versioning(Bucket=bname)["Status"]
        self.assertEqual("Suspended", versioning)

    @functional
    def test_enable_logging(self):
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(
            s3, "S3_AUGMENT_TABLE", [("get_bucket_logging", "Logging", None, None)]
        )
        session_factory = self.replay_flight_data("test_s3_enable_logging")
        bname = "superduper-and-magic"

        session = session_factory()
        client = session.client("s3")
        client.create_bucket(Bucket=bname)
        client.put_bucket_acl(
            Bucket=bname,
            AccessControlPolicy={
                "Owner": {
                    "DisplayName": "mandeep.bal",
                    "ID": "e7c8bb65a5fc49cf906715eae09de9e4bb7861a96361ba79b833aa45f6833b15",
                },
                "Grants": [
                    {
                        "Grantee": {
                            "Type": "Group",
                            "URI": "http://acs.amazonaws.com/groups/s3/LogDelivery",
                        },
                        "Permission": "WRITE",
                    },
                    {
                        "Grantee": {
                            "Type": "Group",
                            "URI": "http://acs.amazonaws.com/groups/s3/LogDelivery",
                        },
                        "Permission": "READ_ACP",
                    },
                ],
            },
        )

        self.addCleanup(destroyBucket, client, bname)

        p = self.load_policy(
            {
                "name": "s3-version",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": [
                    {
                        "type": "toggle-logging",
                        "target_bucket": bname,
                        "target_prefix": "{account}/{source_bucket_name}",
                    }
                ],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 1)
        self.assertEqual(resources[0]["Name"], bname)

        # eventual consistency fun for recording
        if self.recording:
            time.sleep(5)

        logging = client.get_bucket_logging(Bucket=bname).get("LoggingEnabled")
        self.assertTrue(logging)
        self.assertEqual(
            logging["TargetPrefix"], "custodian-skunk-works/superduper-and-magic"
        )

        # Flip the switch
        p = self.load_policy(
            {
                "name": "s3-version",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": [{"type": "toggle-logging", "enabled": False}],
            },
            session_factory=session_factory,
        )

        resources = p.run()
        self.assertEqual(len(resources), 1)

        # eventual consistency fun for recording
        if self.recording:
            time.sleep(12)

        logging = client.get_bucket_logging(Bucket=bname).get("LoggingEnabled")
        self.assertFalse(logging)

    def test_encrypt_policy(self):
        self.patch(
            s3, "S3_AUGMENT_TABLE", [("get_bucket_policy", "Policy", None, "Policy")]
        )
        session_factory = self.replay_flight_data("test_s3_encrypt_policy")
        bname = "custodian-encrypt-test"

        session = session_factory()
        client = session.client("s3")
        client.create_bucket(Bucket=bname)
        self.addCleanup(destroyBucket, client, bname)

        p = self.load_policy(
            {
                "name": "encrypt-keys",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": ["encryption-policy"],
            },
            session_factory=session_factory,
        )
        p.run()
        try:
            resource = session.resource("s3")
            key = resource.Object(bname, "home.txt")
            key.put(Body="hello", ContentLength=5, ContentType="text/plain")
        except ClientError as e:
            self.assertEqual(e.response["Error"]["Code"], "AccessDenied")
        else:
            self.fail("Encryption required policy")

    def test_remove_policy_none_extant(self):
        self.patch(
            s3, "S3_AUGMENT_TABLE", [("get_bucket_policy", "Policy", None, "Policy")]
        )
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        session_factory = self.replay_flight_data("test_s3_remove_empty_policy")
        bname = "custodian-policy-test"
        session = session_factory()
        client = session.client("s3")
        client.create_bucket(Bucket=bname)
        self.addCleanup(destroyBucket, client, bname)
        p = self.load_policy(
            {
                "name": "remove-policy",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": [
                    {"type": "remove-statements", "statement_ids": ["Zebra", "Moon"]}
                ],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 1)
        self.assertRaises(ClientError, client.get_bucket_policy, Bucket=bname)

    def test_remove_policy(self):
        self.patch(
            s3, "S3_AUGMENT_TABLE", [("get_bucket_policy", "Policy", None, "Policy")]
        )
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(s3.RemovePolicyStatement, "executor_factory", MainThreadExecutor)

        session_factory = self.replay_flight_data("test_s3_remove_policy")
        bname = "custodian-policy-test"
        session = session_factory()
        client = session.client("s3")
        client.create_bucket(Bucket=bname)
        client.put_bucket_policy(
            Bucket=bname,
            Policy=json.dumps(
                {
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Sid": "Zebra",
                            "Effect": "Deny",
                            "Principal": "*",
                            "Action": "s3:PutObject",
                            "Resource": "arn:aws:s3:::%s/*" % bname,
                            "Condition": {
                                "StringNotEquals": {
                                    "s3:x-amz-server-side-encryption": [
                                        "AES256", "aws:kms"
                                    ]
                                }
                            },
                        }
                    ],
                }
            ),
        )
        self.addCleanup(destroyBucket, client, bname)
        p = self.load_policy(
            {
                "name": "remove-policy",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": [
                    {"type": "remove-statements", "statement_ids": ["Zebra", "Moon"]}
                ],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 1)
        self.assertRaises(ClientError, client.get_bucket_policy, Bucket=bname)

    def test_remove_policy_matched(self):
        self.patch(
            s3, "S3_AUGMENT_TABLE", [("get_bucket_policy", "Policy", None, "Policy")]
        )
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(s3.RemovePolicyStatement, "executor_factory", MainThreadExecutor)
        self.patch(MainThreadExecutor, "c7n_async", False)

        bname = "custodian-policy-test"
        statement = {
            "Sid": "Zebra",
            "Effect": "Deny",
            "Principal": "*",
            "Action": "s3:PutObject",
            "Resource": "arn:aws:s3:::%s/*" % bname,
            "Condition": {
                "StringNotEquals": {
                    "s3:x-amz-server-side-encryption": ["AES256", "aws:kms"]
                }
            },
        }

        process_buckets = s3.RemovePolicyStatement.process

        def enrich(self, buckets):
            buckets[0]["CrossAccountViolations"] = [statement]
            process_buckets(self, buckets)

        self.patch(s3.RemovePolicyStatement, "process", enrich)

        session_factory = self.replay_flight_data("test_s3_remove_policy")
        session = session_factory()
        client = session.client("s3")
        client.create_bucket(Bucket=bname)
        client.put_bucket_policy(
            Bucket=bname,
            Policy=json.dumps({"Version": "2012-10-17", "Statement": [statement]}),
        )
        self.addCleanup(destroyBucket, client, bname)
        p = self.load_policy(
            {
                "name": "remove-policy",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": [{"type": "remove-statements", "statement_ids": "matched"}],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 1)
        self.assertRaises(ClientError, client.get_bucket_policy, Bucket=bname)

    def test_attach_encrypt_requires_role(self):
        self.assertRaises(
            PolicyValidationError,
            self.load_policy,
            {
                "name": "attach-encrypt",
                "resource": "s3",
                "actions": [{"type": "attach-encrypt"}],
            },
        )

    @skip_if_not_validating
    def test_attach_encrypt_accepts_topic(self):
        p = self.load_policy(
            {
                "name": "attach-encrypt",
                "resource": "s3",
                "actions": [
                    {"type": "attach-encrypt", "role": "-", "topic": "default"}
                ],
            }
        )
        self.assertEqual(p.data["actions"][0]["topic"], "default")

    def test_create_bucket_event(self):
        self.patch(
            s3, "S3_AUGMENT_TABLE", [("get_bucket_policy", "Policy", None, "Policy")]
        )
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        session_factory = self.replay_flight_data("test_s3_create")
        bname = "custodian-create-bucket-v4"
        session = session_factory()
        client = session.client("s3")

        client.create_bucket(Bucket=bname)
        self.addCleanup(destroyBucket, client, bname)

        p = self.load_policy(
            {
                "name": "bucket-create-v2",
                "resource": "s3",
                "mode": {
                    "type": "cloudtrail",
                    "role": "arn:aws:iam::619193117841:role/CustodianDemoRole",
                    "events": ["CreateBucket"],
                },
                "actions": ["encryption-policy"],
            },
            session_factory=session_factory,
        )
        p.push(event_data("event-cloud-trail-create-bucket.json"), None)

        try:
            result = client.get_bucket_policy(Bucket=bname)
        except Exception:
            self.fail("Could not get bucket policy")

        self.assertTrue("Policy" in result)
        policy = json.loads(result["Policy"])
        self.assertEqual(
            policy,
            {
                u"Statement": [
                    {
                        u"Action": u"s3:PutObject",
                        u"Condition": {
                            u"StringNotEquals": {
                                u"s3:x-amz-server-side-encryption": [
                                    u"AES256", u"aws:kms"
                                ]
                            }
                        },
                        u"Effect": u"Deny",
                        u"Principal": u"*",
                        u"Resource": u"arn:aws:s3:::custodian-create-bucket-v4/*",
                        u"Sid": u"RequireEncryptedPutObject",
                    }
                ],
                u"Version": u"2012-10-17",
            },
        )

    def test_attach_encrypt_via_bucket_notification(self):
        self.patch(
            s3, "S3_AUGMENT_TABLE", [("get_bucket_location", "Location", None, None)]
        )
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        session_factory = self.replay_flight_data(
            "test_s3_attach_encrypt_via_bucket_notification"
        )
        bname = "custodian-attach-encrypt-test"
        role = "arn:aws:iam::644160558196:role/custodian-mu"
        self.maxDiff = None
        session = session_factory(region="us-west-2")
        client = session.client("s3")
        client.create_bucket(
            Bucket=bname, CreateBucketConfiguration={"LocationConstraint": "us-west-2"}
        )
        self.addCleanup(destroyBucket, client, bname)

        p = self.load_policy(
            {
                "name": "attach-encrypt",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": [{"type": "attach-encrypt", "role": role}],
            },
            config=Config.empty(region="us-west-2"),
            session_factory=session_factory,
        )

        self.addCleanup(
            LambdaManager(
                functools.partial(session_factory, region="us-west-2")
            ).remove,
            s3crypt.get_function(None, role),
        )

        resources = p.run()
        self.assertEqual(len(resources), 1)
        # time.sleep(10)
        notifications = client.get_bucket_notification_configuration(Bucket=bname)
        notifications.pop("ResponseMetadata")
        self.assertEqual(
            notifications,
            {
                "LambdaFunctionConfigurations": [
                    {
                        "Events": ["s3:ObjectCreated:*"],
                        "Id": "c7n-s3-encrypt",
                        "LambdaFunctionArn": (
                            "arn:aws:lambda:us-west-2:644160558196:function:c7n-s3-encrypt"
                        ),
                    }
                ]
            },
        )
        client.put_object(
            Bucket=bname,
            Key="hello-world.txt",
            Body="hello world",
            ContentType="text/plain",
        )
        # time.sleep(30)
        info = client.head_object(Bucket=bname, Key="hello-world.txt")
        self.assertTrue("ServerSideEncryption" in info)

    def test_attach_encrypt_via_new_topic(self):
        self.patch(
            s3,
            "S3_AUGMENT_TABLE",
            [("get_bucket_notification_configuration", "Notification", None, None)],
        )
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        session_factory = self.replay_flight_data(
            "test_s3_attach_encrypt_via_new_topic"
        )
        bname = "custodian-attach-encrypt-test"
        role = "arn:aws:iam::644160558196:role/custodian-mu"
        self.maxDiff = None
        session = session_factory(region="us-east-1")
        client = session.client("s3")
        client.create_bucket(Bucket=bname)
        self.addCleanup(destroyBucket, client, bname)

        p = self.load_policy(
            {
                "name": "attach-encrypt",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": [
                    {"type": "attach-encrypt", "role": role, "topic": "default"}
                ],
            },
            session_factory=session_factory,
        )

        self.addCleanup(
            LambdaManager(
                functools.partial(session_factory, region="us-east-1")
            ).remove,
            s3crypt.get_function(None, role),
        )
        arn = "arn:aws:sns:us-east-1:644160558196:custodian-attach-encrypt-test"
        self.addCleanup(session.client("sns").delete_topic, TopicArn=arn)
        self.addCleanup(
            session.client("logs").delete_log_group,
            logGroupName="/aws/lambda/c7n-s3-encrypt",
        )

        # Check that the policy sets stuff up properly.
        resources = p.run()
        self.assertEqual(len(resources), 1)
        # time.sleep(10)
        topic_notifications = client.get_bucket_notification_configuration(
            Bucket=bname
        ).get(
            "TopicConfigurations", []
        )
        us = [t for t in topic_notifications if t.get("TopicArn") == arn]
        self.assertEqual(len(us), 1)

        # Check that the stuff behaves properly.
        client.put_object(
            Bucket=bname,
            Key="hello-world.txt",
            Body="hello world",
            ContentType="text/plain",
        )
        # time.sleep(30)
        info = client.head_object(Bucket=bname, Key="hello-world.txt")
        self.assertTrue("ServerSideEncryption" in info)

    def test_attach_encrypt_via_implicit_existing_topic(self):
        self.patch(
            s3,
            "S3_AUGMENT_TABLE",
            [("get_bucket_notification_configuration", "Notification", None, None)],
        )
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        session_factory = self.replay_flight_data(
            "test_s3_attach_encrypt_via_implicit_existing_topic"
        )
        bname = "custodian-attach-encrypt-test"
        role = "arn:aws:iam::644160558196:role/custodian-mu"
        self.maxDiff = None
        session = session_factory(region="us-east-1")
        client = session.client("s3")
        client.create_bucket(Bucket=bname)
        self.addCleanup(destroyBucket, client, bname)

        # Create two sns topics
        topic_configs = []
        for suffix in (".jpg", ".txt"):
            sns = session.client("sns")
            existing_topic_arn = sns.create_topic(
                Name="existing-{}-{}".format(bname, suffix[1:])
            )[
                "TopicArn"
            ]
            policy = {
                "Statement": [
                    {
                        "Action": "SNS:Publish",
                        "Effect": "Allow",
                        "Resource": existing_topic_arn,
                        "Principal": {"Service": "s3.amazonaws.com"},
                    }
                ]
            }
            sns.set_topic_attributes(
                TopicArn=existing_topic_arn,
                AttributeName="Policy",
                AttributeValue=json.dumps(policy),
            )
            self.addCleanup(
                session.client("sns").delete_topic, TopicArn=existing_topic_arn
            )
            topic_configs.append(
                {
                    "TopicArn": existing_topic_arn,
                    "Events": ["s3:ObjectCreated:*"],
                    "Filter": {
                        "Key": {"FilterRules": [{"Name": "suffix", "Value": suffix}]}
                    },
                }
            )
        session.resource("s3").BucketNotification(bname).put(
            NotificationConfiguration={"TopicConfigurations": topic_configs}
        )

        # Now define the policy.
        p = self.load_policy(
            {
                "name": "attach-encrypt",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": [
                    {"type": "attach-encrypt", "role": role, "topic": "default"}
                ],
            },
            session_factory=session_factory,
        )
        self.addCleanup(
            LambdaManager(
                functools.partial(session_factory, region="us-east-1")
            ).remove,
            s3crypt.get_function(None, role),
        )
        self.addCleanup(
            session.client("logs").delete_log_group,
            logGroupName="/aws/lambda/c7n-s3-encrypt",
        )

        # Check that the policy sets stuff up properly.
        resources = p.run()
        self.assertEqual(len(resources), 1)
        # time.sleep(10)
        notifies = client.get_bucket_notification_configuration(Bucket=bname).get(
            "TopicConfigurations", []
        )
        existing = [t for t in notifies if "existing" in t["TopicArn"]]
        self.assertEqual(len(existing), 2)

        # Check that the stuff behaves properly.
        client.put_object(
            Bucket=bname,
            Key="hello-world.txt",
            Body="hello world",
            ContentType="text/plain",
        )
        # time.sleep(30)
        info = client.head_object(Bucket=bname, Key="hello-world.txt")
        self.assertTrue("ServerSideEncryption" in info)

    def test_attach_encrypt_via_explicit_existing_topic(self):
        self.patch(
            s3,
            "S3_AUGMENT_TABLE",
            [("get_bucket_notification_configuration", "Notification", None, None)],
        )
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        session_factory = self.replay_flight_data(
            "test_s3_attach_encrypt_via_explicit_existing_topic"
        )
        bname = "custodian-attach-encrypt-test"
        role = "arn:aws:iam::644160558196:role/custodian-mu"
        self.maxDiff = None
        session = session_factory(region="us-east-1")
        client = session.client("s3")
        client.create_bucket(Bucket=bname)
        self.addCleanup(destroyBucket, client, bname)

        # Create an sns topic
        topic_configs = []
        sns = session.client("sns")
        existing_topic_arn = sns.create_topic(Name="preexisting-{}".format(bname))[
            "TopicArn"
        ]
        policy = {
            "Statement": [
                {
                    "Action": "SNS:Publish",
                    "Effect": "Allow",
                    "Resource": existing_topic_arn,
                    "Principal": {"Service": "s3.amazonaws.com"},
                }
            ]
        }
        sns.set_topic_attributes(
            TopicArn=existing_topic_arn,
            AttributeName="Policy",
            AttributeValue=json.dumps(policy),
        )
        self.addCleanup(session.client("sns").delete_topic, TopicArn=existing_topic_arn)
        topic_configs.append(
            {"TopicArn": existing_topic_arn, "Events": ["s3:ObjectCreated:*"]}
        )
        session.resource("s3").BucketNotification(bname).put(
            NotificationConfiguration={"TopicConfigurations": topic_configs}
        )

        # Now define the policy.
        p = self.load_policy(
            {
                "name": "attach-encrypt",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": [
                    {
                        "type": "attach-encrypt",
                        "role": role,
                        "topic": existing_topic_arn,
                    }
                ],
            },
            session_factory=session_factory,
        )
        self.addCleanup(
            LambdaManager(
                functools.partial(session_factory, region="us-east-1")
            ).remove,
            s3crypt.get_function(None, role),
        )
        self.addCleanup(
            session.client("logs").delete_log_group,
            logGroupName="/aws/lambda/c7n-s3-encrypt",
        )

        # Check that the policy sets stuff up properly.
        resources = p.run()
        self.assertEqual(len(resources), 1)
        # time.sleep(10)
        notifies = client.get_bucket_notification_configuration(Bucket=bname).get(
            "TopicConfigurations", []
        )
        existing = [t for t in notifies if "existing" in t["TopicArn"]]
        self.assertEqual(len(existing), 1)

        # Check that the stuff behaves properly.
        client.put_object(
            Bucket=bname,
            Key="hello-world.txt",
            Body="hello world",
            ContentType="text/plain",
        )
        # time.sleep(30)
        info = client.head_object(Bucket=bname, Key="hello-world.txt")
        self.assertTrue("ServerSideEncryption" in info)

    def test_encrypt_versioned_bucket(self):
        self.patch(
            s3,
            "S3_AUGMENT_TABLE",
            [("get_bucket_versioning", "Versioning", None, None)],
        )

        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(s3.EncryptExtantKeys, "executor_factory", MainThreadExecutor)
        session_factory = self.replay_flight_data("test_s3_encrypt_versioned")
        bname = "custodian-encrypt-test"

        session = session_factory()
        client = session.client("s3")
        client.create_bucket(Bucket=bname)
        client.put_bucket_versioning(
            Bucket=bname, VersioningConfiguration={"Status": "Enabled"}
        )
        self.addCleanup(destroyVersionedBucket, client, bname)
        generateBucketContents(session.resource("s3"), bname)

        p = self.load_policy(
            {
                "name": "encrypt-keys",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": ["encrypt-keys"],
            },
            session_factory=session_factory,
        )
        p.run()
        self.assertTrue(len(client.list_object_versions(Bucket=bname)["Versions"]) == 3)
        self.assertTrue(
            "ServerSideEncryption" in client.head_object(Bucket=bname, Key="home.txt")
        )

    @functional
    def test_encrypt_versioned_bucket_with_existing_keys(self):
        self.patch(
            s3,
            "S3_AUGMENT_TABLE",
            [("get_bucket_versioning", "Versioning", None, None)],
        )

        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(s3.EncryptExtantKeys, "executor_factory", MainThreadExecutor)
        session_factory = self.replay_flight_data(
            "test_s3_encrypt_versioned_bucket_with_existing_keys"
        )
        bname = "custodian-encrypt-test-versioning"

        session = session_factory()
        client = session.client("s3")
        client.create_bucket(Bucket=bname)
        generateBucketContents(
            session.resource("s3"), bname, {"data1.txt": "one", "data2.txt": "two"}
        )
        client.put_bucket_versioning(
            Bucket=bname, VersioningConfiguration={"Status": "Enabled"}
        )
        self.addCleanup(destroyVersionedBucket, client, bname)
        generateBucketContents(session.resource("s3"), bname, {"data1.txt": "three"})

        p = self.load_policy(
            {
                "name": "encrypt-keys",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": ["encrypt-keys"],
            },
            session_factory=session_factory,
        )
        p.run()

        self.assertTrue(len(client.list_object_versions(Bucket=bname)["Versions"]) == 2)
        self.assertTrue(
            "ServerSideEncryption" in client.head_object(Bucket=bname, Key="data1.txt")
        )
        self.assertTrue(
            "ServerSideEncryption" in client.head_object(Bucket=bname, Key="data2.txt")
        )

    def test_encrypt_key_empty_bucket(self):
        self.patch(s3, "S3_AUGMENT_TABLE", [])
        self.patch(s3.EncryptExtantKeys, "executor_factory", MainThreadExecutor)
        session_factory = self.replay_flight_data("test_s3_encrypt_empty")
        bname = "custodian-encrypt-test"

        session = session_factory()
        client = session.client("s3")
        client.create_bucket(Bucket=bname)
        self.addCleanup(destroyBucket, client, bname)

        p = self.load_policy(
            {
                "name": "encrypt-keys",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": ["encrypt-keys"],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 1)

    def test_encrypt_keys(self):
        self.patch(s3, "S3_AUGMENT_TABLE", [])
        session_factory = self.replay_flight_data("test_s3_encrypt")
        bname = "custodian-encrypt-test"

        session = session_factory()
        client = session.client("s3")
        client.create_bucket(Bucket=bname)
        self.addCleanup(destroyBucket, client, bname)
        generateBucketContents(session.resource("s3"), bname)

        # start with a report-only option since it doesn't modify the bucket
        report_policy = self.load_policy(
            {
                "name": "encrypt-keys",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": [{"type": "encrypt-keys", "report-only": True}],
            },
            session_factory=session_factory,
        )
        report_resources = report_policy.run()

        self.assertEqual(report_resources[0]["KeyRemediated"], 3)

        p = self.load_policy(
            {
                "name": "encrypt-keys",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": ["encrypt-keys"],
            },
            session_factory=session_factory,
        )
        p.run()

        self.assertTrue(
            "ServerSideEncryption" in client.head_object(Bucket=bname, Key="home.txt")
        )

        # re-run the report policy after to ensure we have no items
        # needing remediation
        report_resources = report_policy.run()
        self.assertEqual(report_resources[0]["KeyRemediated"], 0)

    def test_encrypt_keys_aes256_sufficient(self):
        self.patch(s3, "S3_AUGMENT_TABLE", [])
        session_factory = self.replay_flight_data("test_s3_encrypt_aes256_sufficient")
        bname = "custodian-encrypt-sufficient-test"

        session = session_factory()
        client = session.client("s3")
        kms = session.client("kms")

        client.create_bucket(Bucket=bname)
        self.addCleanup(destroyBucket, client, bname)
        key_id = [
            k
            for k in kms.list_aliases().get("Aliases", ())
            if k["AliasName"] == "alias/aws/s3"
        ][
            0
        ][
            "AliasArn"
        ]

        client.put_object(
            Bucket=bname,
            Key="testing-abc",
            ServerSideEncryption="aws:kms",
            SSEKMSKeyId=key_id,
        )
        client.put_object(
            Bucket=bname, Key="testing-123", ServerSideEncryption="AES256"
        )

        p = self.load_policy(
            {
                "name": "encrypt-keys",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": [{"type": "encrypt-keys"}],
            },
            session_factory=session_factory,
        )

        p.run()

        result = client.head_object(Bucket=bname, Key="testing-123")
        self.assertTrue(result["ServerSideEncryption"] == "AES256")

        result = client.head_object(Bucket=bname, Key="testing-abc")
        self.assertTrue(result["ServerSideEncryption"] == "aws:kms")
        data = json.load(
            open(os.path.join(p.ctx.output.root_dir, "action-encryptextantkeys"))
        )
        self.assertEqual([{"Count": 2, "Remediated": 0, "Bucket": bname}], data)

    def test_encrypt_keys_key_id_option(self):
        self.patch(s3, "S3_AUGMENT_TABLE", [])
        session_factory = self.replay_flight_data("test_s3_encrypt_key_id_option")
        bname = "custodian-encrypt-test"

        session = session_factory()
        client = session.client("s3")
        client.create_bucket(Bucket=bname)
        self.addCleanup(destroyBucket, client, bname)
        generateBucketContents(session.resource("s3"), bname)

        key_one = "845ab6f1-744c-4edc-b702-efae6836818a"
        p = self.load_policy(
            {
                "name": "encrypt-keys",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": [
                    {"type": "encrypt-keys", "crypto": "aws:kms", "key-id": key_one}
                ],
            },
            session_factory=session_factory,
        )
        p.run()
        result = client.head_object(Bucket=bname, Key="home.txt")
        self.assertTrue("SSEKMSKeyId" in result)
        self.assertTrue(key_one in result["SSEKMSKeyId"])

        # Now test that we can re-key it to something else
        key_two = "5fd9f6d6-4294-4926-8719-1e85695e2ad6"
        p = self.load_policy(
            {
                "name": "encrypt-keys",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": [
                    {"type": "encrypt-keys", "crypto": "aws:kms", "key-id": key_two}
                ],
            },
            session_factory=session_factory,
        )
        p.run()
        result = client.head_object(Bucket=bname, Key="home.txt")
        self.assertTrue("SSEKMSKeyId" in result)
        self.assertTrue(key_two in result["SSEKMSKeyId"])

    def test_global_grants_filter_option(self):
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(s3, "S3_AUGMENT_TABLE", [("get_bucket_acl", "Acl", None, None)])
        session_factory = self.replay_flight_data("test_s3_global_grants_filter")
        bname = "custodian-testing-grants"
        session = session_factory()
        client = session.client("s3")
        client.create_bucket(Bucket=bname)
        self.addCleanup(destroyBucket, client, bname)

        public = "http://acs.amazonaws.com/groups/global/AllUsers"

        client.put_bucket_acl(
            Bucket=bname,
            AccessControlPolicy={
                "Owner": {
                    "DisplayName": "k_vertigo",
                    "ID": "904fc4c4790937100e9eb293a15e6a0a1f265a064888055b43d030034f8881ee",
                },
                "Grants": [
                    {"Grantee": {"Type": "Group", "URI": public}, "Permission": "WRITE"}
                ],
            },
        )
        p = self.load_policy(
            {
                "name": "s3-global-check",
                "resource": "s3",
                "filters": [
                    {"Name": "custodian-testing-grants"},
                    {"type": "global-grants", "permissions": ["READ_ACP"]},
                ],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 0)

        p = self.load_policy(
            {
                "name": "s3-global-check",
                "resource": "s3",
                "filters": [
                    {"Name": "custodian-testing-grants"},
                    {"type": "global-grants", "permissions": ["WRITE"]},
                ],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 1)

    def test_global_grants_filter_and_remove(self):
        self.patch(s3, "S3_AUGMENT_TABLE", [("get_bucket_acl", "Acl", None, None)])
        session_factory = self.replay_flight_data("test_s3_grants")

        bname = "custodian-testing-grants"
        session = session_factory()
        client = session.client("s3")
        client.create_bucket(Bucket=bname)

        public = "http://acs.amazonaws.com/groups/global/AllUsers"
        client.put_bucket_acl(
            Bucket=bname,
            AccessControlPolicy={
                "Owner": {
                    "DisplayName": "k_vertigo",
                    "ID": "904fc4c4790937100e9eb293a15e6a0a1f265a064888055b43d030034f8881ee",
                },
                "Grants": [
                    {"Grantee": {"Type": "Group", "URI": public}, "Permission": "WRITE"}
                ],
            },
        )
        p = self.load_policy(
            {
                "name": "s3-remove-global",
                "resource": "s3",
                "filters": [
                    {"Name": "custodian-testing-grants"}, {"type": "global-grants"}
                ],
                "actions": [{"type": "delete-global-grants", "grantees": [public]}],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        grants = client.get_bucket_acl(Bucket=bname)
        client.delete_bucket(Bucket=bname)
        self.assertEqual(grants["Grants"], [])
        self.assertEqual(resources[0]["Name"], bname)

    def test_s3_mark_for_op(self):
        self.patch(
            s3, "S3_AUGMENT_TABLE", [("get_bucket_tagging", "Tags", [], "TagSet")]
        )
        session_factory = self.replay_flight_data("test_s3_mark_for_op")
        session = session_factory()
        client = session.client("s3")
        bname = "custodian-mark-test"
        p = self.load_policy(
            {
                "name": "s3-mark",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": [{"type": "mark-for-op", "days": 3, "op": "delete"}],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 1)
        tags = client.get_bucket_tagging(Bucket=bname)
        tag_map = {t["Key"]: t["Value"] for t in tags.get("TagSet", {})}
        self.assertTrue("maid_status" in tag_map)
        self.assertTrue("delete" in tag_map.get("maid_status"))

    def test_s3_remove_tag(self):
        self.patch(
            s3, "S3_AUGMENT_TABLE", [("get_bucket_tagging", "Tags", [], "TagSet")]
        )
        self.patch(s3.RemoveTag, "executor_factory", MainThreadExecutor)
        session_factory = self.replay_flight_data("test_s3_remove_tag")
        session = session_factory()
        client = session.client("s3")
        bname = "custodian-mark-test"
        p = self.load_policy(
            {
                "name": "s3-unmark",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": ["unmark"],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 1)
        tags = client.get_bucket_tagging(Bucket=bname)
        tag_map = {t["Key"]: t["Value"] for t in tags.get("TagSet", {})}
        self.assertTrue("maid_status" not in tag_map)
        old_tags = {t["Key"]: t["Value"] for t in resources[0]["Tags"]}
        self.assertTrue("maid_status" in old_tags)

    def test_hosts_website(self):
        self.patch(
            s3, "S3_AUGMENT_TABLE", [("get_bucket_website", "Website", None, None)]
        )
        session_factory = self.replay_flight_data("test_s3_hosts_website")
        session = session_factory()
        client = session.client("s3")
        bname = "custodian-static-website-test"
        client.create_bucket(Bucket=bname)
        client.put_bucket_website(
            Bucket=bname,
            WebsiteConfiguration={
                "ErrorDocument": {"Key": "error.html"},
                "IndexDocument": {"Suffix": "index.html"},
            },
        )
        self.addCleanup(client.delete_bucket, Bucket=bname)
        p = self.load_policy(
            {
                "name": "s3-website-hosting",
                "resource": "s3",
                "filters": [{"Website": "not-null"}],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 1)
        names = [b["Name"] for b in resources]
        self.assertTrue(bname in names)

        p = self.load_policy(
            {
                "name": "s3-website-hosting",
                "resource": "s3",
                "filters": [{"Website": "not-null"}],
                "actions": ["remove-website-hosting"],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 0)

    def test_delete_bucket_notification(self):
        self.patch(
            s3,
            "S3_AUGMENT_TABLE",
            [("get_bucket_notification_configuration", "Notification", None, None)],
        )
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        session_factory = self.replay_flight_data("test_s3_delete_bucket_notification")
        bname = "custodian-delete-bucket-notification-test"
        config_id = "c7n-notify-1"
        self.maxDiff = None
        session = session_factory(region="us-east-1")
        client = session.client("s3")
        client.create_bucket(Bucket=bname)
        self.addCleanup(destroyBucket, client, bname)

        p = self.load_policy(
            {
                "name": "s3-delete-bucket-notification",
                "resource": "s3",
                "filters": [
                    {"Name": bname},
                    {
                        "type": "bucket-notification",
                        "kind": "sns",
                        "key": "Id",
                        "value": config_id,
                        "op": "eq",
                    },
                ],
                "actions": [
                    {"type": "delete-bucket-notification", "statement_ids": "matched"}
                ],
            },
            session_factory=session_factory,
        )

        topic_arn = session.client("sns").create_topic(Name="bucket-notification-test")[
            "TopicArn"
        ]
        self.addCleanup(session.client("sns").delete_topic, TopicArn=topic_arn)
        topic_policy = {
            "Statement": [
                {
                    "Action": "SNS:Publish",
                    "Effect": "Allow",
                    "Resource": topic_arn,
                    "Principal": {"Service": "s3.amazonaws.com"},
                }
            ]
        }
        session.client("sns").set_topic_attributes(
            TopicArn=topic_arn,
            AttributeName="Policy",
            AttributeValue=json.dumps(topic_policy),
        )
        client.put_bucket_notification_configuration(
            Bucket=bname,
            NotificationConfiguration={
                "TopicConfigurations": [
                    {
                        "TopicArn": topic_arn,
                        "Events": ["s3:ObjectCreated:*"],
                        "Id": config_id,
                    },
                    {
                        "TopicArn": topic_arn,
                        "Events": ["s3:ObjectRemoved:*"],
                        "Id": "another1",
                    },
                ]
            },
        )

        resources = p.run()
        self.assertEqual(len(resources), 1)
        # time.sleep(10)
        topic_notifications = client.get_bucket_notification_configuration(
            Bucket=bname
        ).get(
            "TopicConfigurations", []
        )
        us = [t for t in topic_notifications if t.get("TopicArn") == topic_arn]
        self.assertEqual(len(us), 1)

    def test_enable_bucket_encryption_kms(self):
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(s3, "S3_AUGMENT_TABLE", [])
        session_factory = self.replay_flight_data(
            "test_s3_enable_bucket_encryption_kms"
        )
        session = session_factory()
        client = session.client("s3")
        kms_client = session.client("kms")
        bname = "custodian-enable-bucket-encryption-kms"

        client.create_bucket(Bucket=bname)
        self.addCleanup(destroyBucket, client, bname)

        with self.assertRaises(Exception):
            response = client.get_bucket_encryption(Bucket=bname)

        key = kms_client.list_keys()["Keys"][0]
        key_arn = kms_client.describe_key(KeyId=key["KeyId"])["KeyMetadata"]["Arn"]
        p = self.load_policy(
            {
                "name": "s3-enable-bucket-encryption",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": [
                    {
                        "type": "set-bucket-encryption",
                        "key": str(key["KeyId"]),
                        "crypto": "aws:kms",
                    }
                ],
            },
            session_factory=session_factory,
        )

        resources = p.run()
        self.assertEqual(len(resources), 1)

        if self.recording:
            time.sleep(5)

        response = client.get_bucket_encryption(Bucket=bname)
        rules = response["ServerSideEncryptionConfiguration"]["Rules"][0][
            "ApplyServerSideEncryptionByDefault"
        ]
        self.assertEqual(rules["SSEAlgorithm"], "aws:kms")
        self.assertEqual(rules["KMSMasterKeyID"], key_arn)

    def test_enable_bucket_encryption_kms_alias(self):
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(s3, "S3_AUGMENT_TABLE", [])
        session_factory = self.replay_flight_data(
            "test_s3_enable_bucket_encryption_kms_alias"
        )
        session = session_factory()
        client = session.client("s3")
        kms_client = session.client("kms")
        bname = "custodian-enable-bucket-encryption-kms-alias"
        client.create_bucket(Bucket=bname)
        self.addCleanup(destroyBucket, client, bname)

        kms_alias = "alias/some-key"
        kms_alias_id = kms_client.describe_key(KeyId=kms_alias)["KeyMetadata"]["Arn"]
        p = self.load_policy(
            {
                "name": "s3-enable-bucket-encryption-alias",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": [
                    {
                        "type": "set-bucket-encryption",
                        "crypto": "aws:kms",
                        "key": kms_alias,
                    }
                ],
            },
            session_factory=session_factory,
        )

        resources = p.run()
        self.assertEqual(len(resources), 1)

        if self.recording:
            time.sleep(5)

        response = client.get_bucket_encryption(Bucket=bname)
        rules = response["ServerSideEncryptionConfiguration"]["Rules"][0][
            "ApplyServerSideEncryptionByDefault"
        ]
        self.assertEqual(rules["SSEAlgorithm"], "aws:kms")
        self.assertEqual(rules["KMSMasterKeyID"], kms_alias_id)

        p = self.load_policy(
            {
                "name": "s3-enable-bucket-encryption-bad-alias",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": [
                    {
                        "type": "set-bucket-encryption",
                        "crypto": "aws:kms",
                        "key": "alias/some-nonexistant-alias",
                    }
                ],
            },
            session_factory=session_factory,
        )

        resources = p.run()
        self.assertEqual(len(resources), 1)

        if self.recording:
            time.sleep(5)

        response = client.get_bucket_encryption(Bucket=bname)
        rules = response["ServerSideEncryptionConfiguration"]["Rules"][0][
            "ApplyServerSideEncryptionByDefault"
        ]
        self.assertEqual(rules["SSEAlgorithm"], "aws:kms")
        self.assertIsNone(rules.get("KMSMasterKeyID"))

    def test_enable_bucket_encryption_aes256(self):
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(s3, "S3_AUGMENT_TABLE", [])
        session_factory = self.replay_flight_data(
            "test_s3_enable_bucket_encryption_aes256"
        )
        session = session_factory()
        client = session.client("s3")
        bname = "custodian-enable-bucket-encryption-aes256"

        client.create_bucket(Bucket=bname)
        self.addCleanup(destroyBucket, client, bname)

        with self.assertRaises(Exception):
            client.get_bucket_encryption(Bucket=bname)

        p = self.load_policy(
            {
                "name": "s3-enable-bucket-encryption",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": [{"type": "set-bucket-encryption"}],
            },
            session_factory=session_factory,
        )

        resources = p.run()
        self.assertEqual(len(resources), 1)
        if self.recording:
            time.sleep(5)

        response = client.get_bucket_encryption(Bucket=bname)
        rules = response["ServerSideEncryptionConfiguration"]["Rules"][0][
            "ApplyServerSideEncryptionByDefault"
        ]
        self.assertEqual(rules["SSEAlgorithm"], "AES256")

        client.delete_bucket_encryption(Bucket=bname)
        if self.recording:
            time.sleep(5)

        with self.assertRaises(Exception):
            client.get_bucket_encryption(Bucket=bname)

        p = self.load_policy(
            {
                "name": "s3-enable-bucket-encryption",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": [{"type": "set-bucket-encryption", "crypto": "AES256"}],
            },
            session_factory=session_factory,
        )

        resources = p.run()
        self.assertEqual(len(resources), 1)
        if self.recording:
            time.sleep(5)

        response = client.get_bucket_encryption(Bucket=bname)
        rules = response["ServerSideEncryptionConfiguration"]["Rules"][0][
            "ApplyServerSideEncryptionByDefault"
        ]
        self.assertEqual(rules["SSEAlgorithm"], "AES256")

    def test_delete_bucket_encryption(self):
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(s3, "S3_AUGMENT_TABLE", [])
        session_factory = self.replay_flight_data("test_s3_delete_bucket_encryption")
        session = session_factory()
        client = session.client("s3")
        bname = "custodian-delete-bucket-encryption-aes256"

        client.create_bucket(Bucket=bname)
        self.addCleanup(destroyBucket, client, bname)

        with self.assertRaises(Exception):
            client.get_bucket_encryption(Bucket=bname)

        client.put_bucket_encryption(
            Bucket=bname,
            ServerSideEncryptionConfiguration={
                "Rules": [
                    {"ApplyServerSideEncryptionByDefault": {"SSEAlgorithm": "AES256"}}
                ]
            },
        )

        p = self.load_policy(
            {
                "name": "s3-delete-bucket-encryption",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": [{"type": "set-bucket-encryption", "enabled": False}],
            },
            session_factory=session_factory,
        )

        resources = p.run()
        self.assertEqual(len(resources), 1)
        if self.recording:
            time.sleep(5)
        with self.assertRaises(Exception):
            client.get_bucket_encryption(Bucket=bname)


class S3LifecycleTest(BaseTest):

    def test_lifecycle(self):
        self.patch(s3.S3, "executor_factory", MainThreadExecutor)
        self.patch(
            s3,
            "S3_AUGMENT_TABLE",
            [("get_bucket_lifecycle_configuration", "Lifecycle", None, None)],
        )
        session_factory = self.replay_flight_data("test_s3_lifecycle")
        session = session_factory()
        client = session.client("s3")
        bname = "custodian-lifecycle-test"

        # Make a bucket
        client.create_bucket(Bucket=bname)
        self.addCleanup(destroyBucket, client, bname)
        buckets = set([b["Name"] for b in client.list_buckets()["Buckets"]])
        self.assertIn(bname, buckets)

        def get_policy(**kwargs):
            rule = {
                "Status": "Enabled",
                "Prefix": "foo/",
                "Transitions": [{"Days": 60, "StorageClass": "GLACIER"}],
            }
            rule.update(**kwargs)

            policy = {
                "name": "s3-lifecycle",
                "resource": "s3",
                "filters": [{"Name": bname}],
                "actions": [{"type": "configure-lifecycle", "rules": [rule]}],
            }
            return policy

        def run_policy(policy):
            p = self.load_policy(policy, session_factory=session_factory)
            resources = p.run()
            self.assertEqual(len(resources), 1)

            if self.recording:
                time.sleep(5)

        #
        # Add the first lifecycle
        #
        lifecycle_id1 = "test-lifecycle"
        policy = get_policy(ID=lifecycle_id1)
        run_policy(policy)
        lifecycle = client.get_bucket_lifecycle_configuration(Bucket=bname)
        self.assertEqual(lifecycle["Rules"][0]["ID"], lifecycle_id1)

        #
        # Now add another lifecycle rule to ensure it doesn't clobber the first one
        #
        lifecycle_id2 = "test-lifecycle-two"
        policy = get_policy(ID=lifecycle_id2, Prefix="bar/")
        run_policy(policy)

        # Verify the lifecycle
        lifecycle = client.get_bucket_lifecycle_configuration(Bucket=bname)
        self.assertEqual(len(lifecycle["Rules"]), 2)
        self.assertSetEqual(
            set([x["ID"] for x in lifecycle["Rules"]]),
            set([lifecycle_id1, lifecycle_id2]),
        )

        #
        # Next, overwrite one of the lifecycles and make sure it changed
        #
        policy = get_policy(ID=lifecycle_id2, Prefix="baz/")
        run_policy(policy)

        # Verify the lifecycle
        lifecycle = client.get_bucket_lifecycle_configuration(Bucket=bname)
        self.assertEqual(len(lifecycle["Rules"]), 2)
        self.assertSetEqual(
            set([x["ID"] for x in lifecycle["Rules"]]),
            set([lifecycle_id1, lifecycle_id2]),
        )

        for rule in lifecycle["Rules"]:
            if rule["ID"] == lifecycle_id2:
                self.assertEqual(rule["Prefix"], "baz/")

        #
        # Test deleting a lifecycle
        #
        policy = get_policy(ID=lifecycle_id1, Status="absent")
        run_policy(policy)

        lifecycle = client.get_bucket_lifecycle_configuration(Bucket=bname)
        self.assertEqual(len(lifecycle["Rules"]), 1)
        self.assertEqual(lifecycle["Rules"][0]["ID"], lifecycle_id2)



# Copyright 2016-2017 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import, division, print_function, unicode_literals

import json

from c7n.actions import RemovePolicyBase, ModifyPolicyBase, BaseAction
from c7n.filters import CrossAccountAccessFilter, PolicyChecker
from c7n.filters.kms import KmsRelatedFilter
from c7n.manager import resources
from c7n.query import QueryResourceManager
from c7n.resolver import ValuesFrom
from c7n.utils import local_session, type_schema


@resources.register('sns')
class SNS(QueryResourceManager):

    class resource_type(object):
        service = 'sns'
        type = 'topic'
        enum_spec = ('list_topics', 'Topics', None)
        detail_spec = (
            'get_topic_attributes', 'TopicArn', 'TopicArn', 'Attributes')
        id = 'TopicArn'
        filter_name = None
        filter_type = None
        name = 'DisplayName'
        date = None
        dimension = 'TopicName'
        default_report_fields = (
            'TopicArn',
            'DisplayName',
            'SubscriptionsConfirmed',
            'SubscriptionsPending',
            'SubscriptionsDeleted'
        )


class SNSPolicyChecker(PolicyChecker):

    @property
    def allowed_endpoints(self):
        return self.checker_config.get('allowed_endpoints', ())

    @property
    def allowed_protocols(self):
        return self.checker_config.get('allowed_protocols', ())

    def handle_sns_endpoint(self, s, c):
        conditions = self.normalize_conditions(s)
        # yield to aws:sourceowner
        if not self.allowed_endpoints:
            return not any(
                single_condition.get('key', None) == 'aws:sourceowner'
                for single_condition in conditions
            )
        # check if any of the allowed_endpoints are a substring
        # to any of the values in the condition
        for value in c['values']:
            if not any(endpoint in value for endpoint in self.allowed_endpoints):
                return True
        return False

    def handle_sns_protocol(self, s, c):
        return bool(set(c['values']).difference(self.allowed_protocols))


@SNS.filter_registry.register('cross-account')
class SNSCrossAccount(CrossAccountAccessFilter):
    """Filter to return all SNS topics with cross account access permissions

    The whitelist parameter will omit the accounts that match from the return

    :example:

        .. code-block:

            policies:
              - name: sns-cross-account
                resource: sns
                filters:
                  - type: cross-account
                    whitelist:
                      - permitted-account-01
                      - permitted-account-02
    """

    valid_protocols = (
        "http",
        "https",
        "email",
        "email-json",
        "sms",
        "sqs",
        "application",
        "lambda"
    )

    schema = type_schema(
        'cross-account',
        rinherit=CrossAccountAccessFilter.schema,
        whitelist_endpoints={'type': 'array', 'items': {'type': 'string'}},
        whitelist_endpoints_from=ValuesFrom.schema,
        whitelist_protocols={'type': 'array', 'items': {'type': 'string', 'enum': valid_protocols}},
        whitelist_protocols_from=ValuesFrom.schema
    )

    permissions = ('sns:GetTopicAttributes',)

    checker_factory = SNSPolicyChecker

    def process(self, resources, event=None):
        self.endpoints = self.get_endpoints()
        self.protocols = self.get_protocols()
        self.checker_config = getattr(self, 'checker_config', None) or {}
        self.checker_config.update(
            {
                'allowed_endpoints': self.endpoints,
                'allowed_protocols': self.protocols
            }
        )
        return super(SNSCrossAccount, self).process(resources, event)

    def get_endpoints(self):
        endpoints = set(self.data.get('whitelist_endpoints', ()))
        if 'whitelist_endpoints_from' in self.data:
            values = ValuesFrom(self.data['whitelist_endpoints_from'], self.manager)
            endpoints = endpoints.union(values.get_values())
        return endpoints

    def get_protocols(self):
        protocols = set(self.data.get('whitelist_protocols', ()))
        if 'whitelist_protocols_from' in self.data:
            values = ValuesFrom(self.data['whitelist_protocols_from'], self.manager)
            protocols = protocols.union(
                [p for p in values.get_values() if p in self.valid_protocols]
            )
        return protocols


@SNS.action_registry.register('remove-statements')
class RemovePolicyStatement(RemovePolicyBase):
    """Action to remove policy statements from SNS

    :example:

    .. code-block:: yaml

           policies:
              - name: sns-cross-account
                resource: sns
                filters:
                  - type: cross-account
                actions:
                  - type: remove-statements
                    statement_ids: matched
    """

    permissions = ('sns:SetTopicAttributes', 'sns:GetTopicAttributes')

    def process(self, resources):
        results = []
        client = local_session(self.manager.session_factory).client('sns')
        for r in resources:
            try:
                results += filter(None, [self.process_resource(client, r)])
            except Exception:
                self.log.exception(
                    "Error processing sns:%s", r['TopicArn'])
        return results

    def process_resource(self, client, resource):
        p = resource.get('Policy')
        if p is None:
            return

        p = json.loads(resource['Policy'])
        statements, found = self.process_policy(
            p, resource, CrossAccountAccessFilter.annotation_key)

        if not found:
            return

        client.set_topic_attributes(
            TopicArn=resource['TopicArn'],
            AttributeName='Policy',
            AttributeValue=json.dumps(p)
        )
        return {'Name': resource['TopicArn'],
                'State': 'PolicyRemoved',
                'Statements': found}


@SNS.action_registry.register('modify-policy')
class ModifyPolicyStatement(ModifyPolicyBase):
    """Action to modify policy statements from SNS

    :example:

    .. code-block:: yaml

           policies:
              - name: sns-cross-account
                resource: sns
                filters:
                  - type: cross-account
                actions:
                  - type: modify-policy
                    add-statements: [statement]
                    remove-statements: [statement_id] or * or 'matched'
    """

    permissions = ('sns:SetTopicAttributes', 'sns:GetTopicAttributes')

    def process(self, resources):
        results = []
        client = local_session(self.manager.session_factory).client('sns')
        for r in resources:
            policy = json.loads(r.get('Policy') or '{}')
            policy_statements = policy.setdefault('Statement', [])

            new_policy, removed = self.remove_statements(
                policy_statements, r, CrossAccountAccessFilter.annotation_key)
            if new_policy is None:
                new_policy = policy_statements
            new_policy, added = self.add_statements(new_policy)

            if not removed or not added:
                continue

            results += {
                'Name': r['TopicArn'],
                'State': 'PolicyModified',
                'Statements': new_policy
            }
            policy['Statement'] = new_policy
            client.set_topic_attributes(
                TopicArn=r['TopicArn'],
                AttributeName='Policy',
                AttributeValue=json.dumps(policy)
            )
        return results


@SNS.filter_registry.register('kms-key')
class KmsFilter(KmsRelatedFilter):

    RelatedIdsExpression = 'KmsMasterKeyId'


@SNS.action_registry.register('set-encryption')
class SetEncryption(BaseAction):
    """
    Set Encryption on SNS Topics

    By default if no key is specified, alias/aws/sns is used

    key can either be a KMS key ARN, key id, or an alias

    :example:

    .. code-block:: yaml

        policies:
            - name: set-sns-topic-encryption
              resource: sns
              actions:
                - type: set-encryption
                  key: alias/cmk/key
                  enabled: True

            - name: set-sns-topic-encryption-with-id
              resource: sns
              actions:
                - type: set-encryption
                  key: abcdefgh-1234-1234-1234-123456789012
                  enabled: True

            - name: set-sns-topic-encryption-with-arn
              resource: sns
              actions:
                - type: set-encryption
                  key: arn:aws:kms:us-west-1:123456789012:key/abcdefgh-1234-1234-1234-123456789012
                  enabled: True
    """

    schema = type_schema(
        'set-encryption',
        enabled={'type': 'boolean'},
        key={'type': 'string'}
    )

    permissions = ('sns:SetTopicAttributes', 'kms:DescribeKey',)

    def process(self, resources):
        sns = local_session(self.manager.session_factory).client('sns')

        if self.data.get('enabled', True):
            key = self.data.get('key', 'alias/aws/sns')
        else:
            key = ''

        for r in resources:
            sns.set_topic_attributes(
                TopicArn=r['TopicArn'],
                AttributeName='KmsMasterKeyId',
                AttributeValue=key
            )
        return resources



# Copyright 2016-2017 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import, division, print_function, unicode_literals

import datetime
from dateutil.parser import parse as date_parse

from .common import BaseTest
from .test_offhours import mock_datetime_now
from time import sleep


class ElasticBeanstalkEnvironment(BaseTest):

    def test_resource_manager(self):
        factory = self.replay_flight_data("test_elasticbeanstalk_describe_envs")
        p = self.load_policy(
            {"name": "eb-env-query", "resource": "elasticbeanstalk-environment"},
            session_factory=factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 2)

    def test_eb_env_regex(self):
        factory = self.replay_flight_data("test_elasticbeanstalk_describe_envs")
        p = self.load_policy(
            {
                "name": "eb-find-inactive",
                "resource": "elasticbeanstalk-environment",
                "filters": [
                    {
                        "type": "value",
                        "key": "CNAME",
                        "op": "regex",
                        "value": ".*inactive.*",
                    }
                ],
            },
            session_factory=factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 1)

    def test_eb_env_uptime(self):
        factory = self.replay_flight_data("test_elasticbeanstalk_describe_envs")
        p = self.load_policy(
            {
                "name": "eb-find-inactive",
                "resource": "elasticbeanstalk-environment",
                "filters": [
                    {
                        "type": "value",
                        "key": "DateCreated",
                        "value": 1,
                        "value_type": "age",
                        "op": "greater-than",
                    }
                ],
            },
            session_factory=factory,
        )
        with mock_datetime_now(date_parse("2017-12-19"), datetime):
            resources = p.run()
        self.assertEqual(len(resources), 2)


class EbEnvBaseTest(BaseTest):

    def query_env_status(self, session, env_name):
        client = session.client("elasticbeanstalk")
        res = client.describe_environments(EnvironmentNames=[env_name])
        if len(res["Environments"]) > 0:
            return res["Environments"][0]["Status"]
        return None

    def env_tags_dict(self, session, env_arn):
        client = session.client("elasticbeanstalk")
        tagres = client.list_tags_for_resource(ResourceArn=env_arn)
        tags = tagres["ResourceTags"]
        return {t["Key"]: t["Value"] for t in tags}


class TestTerminate(EbEnvBaseTest):

    def test_eb_env_terminate(self):
        envname = "c7n-eb-tag-test-inactive"
        session_factory = self.replay_flight_data("test_eb_env_terminate")
        assert self.query_env_status(session_factory(), envname) == "Ready"
        p = self.load_policy(
            {
                "name": "eb-env-term",
                "resource": "elasticbeanstalk-environment",
                "filters": [{"EnvironmentName": envname}],
                "actions": [{"type": "terminate"}],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 1)
        assert resources[0]["EnvironmentName"] == envname
        assert self.query_env_status(session_factory(), envname) == "Terminating"


class TestEBEnvTagging(EbEnvBaseTest):

    def test_tag_delayed(self):
        envname = "c7n-eb-tag-test-inactive"
        envarn = ("arn:aws:elasticbeanstalk:us-east-1:012345678901:"
                  "environment/re-jenkins/%s" % envname)
        factory = self.replay_flight_data("test_elasticbeanstalk_env_tag_delayed")
        p = self.load_policy(
            {
                "name": "eb-tag-delayed",
                "resource": "elasticbeanstalk-environment",
                "filters": [
                    {
                        "type": "value",
                        "key": "CNAME",
                        "op": "regex",
                        "value": ".*inactive.*",
                    }
                ],
                "actions": [
                    {
                        "type": "mark-for-op",
                        "op": "terminate",
                        "days": 7,
                        "tag": "c7n-eb-tag-test",
                    }
                ],
            },
            session_factory=factory,
        )
        if self.recording:
            resources = p.run()
        else:
            with mock_datetime_now(date_parse("2017-11-10"), datetime):
                resources = p.run()
        self.assertEqual(len(resources), 1)
        self.assertEqual(resources[0]["EnvironmentName"], envname)
        if self.recording:
            sleep(4)
        while self.query_env_status(factory(), envname) != "Ready":
            if self.recording:
                sleep(30)
            pass
        self.assertEqual(
            self.env_tags_dict(factory(), envarn).get("c7n-eb-tag-test"),
            "Resource does not meet policy: terminate@2017/12/12",
        )

    def test_tag(self):
        envname = "c7n-eb-tag-test-inactive"
        envarn = ("arn:aws:elasticbeanstalk:us-east-1:012345678901:"
                  "environment/re-jenkins/%s" % envname)
        factory = self.replay_flight_data("test_elasticbeanstalk_env_tag")
        p = self.load_policy(
            {
                "name": "eb-tag",
                "resource": "elasticbeanstalk-environment",
                "filters": [
                    {
                        "type": "value",
                        "key": "CNAME",
                        "op": "regex",
                        "value": ".*inactive.*",
                    }
                ],
                "actions": [
                    {"type": "tag", "key": "tagTestKey", "value": "tagTestValue"}
                ],
            },
            session_factory=factory,
        )
        if self.recording:
            resources = p.run()
        else:
            with mock_datetime_now(date_parse("2017-11-10"), datetime):
                resources = p.run()
        self.assertEqual(len(resources), 1)
        self.assertEqual(resources[0]["EnvironmentName"], envname)
        if self.recording:
            sleep(5)
        while self.query_env_status(factory(), envname) != "Ready":
            if self.recording:
                sleep(30)
            pass
        self.assertEqual(
            self.env_tags_dict(factory(), envarn).get("tagTestKey"), "tagTestValue"
        )

    def test_unmark(self):
        envname = "c7n-eb-tag-test-inactive"
        envarn = ("arn:aws:elasticbeanstalk:us-east-1:012345678901:"
                  "environment/re-jenkins/%s" % envname)
        factory = self.replay_flight_data("test_elasticbeanstalk_env_unmark")
        p = self.load_policy(
            {
                "name": "eb-tag",
                "resource": "elasticbeanstalk-environment",
                "filters": [
                    {
                        "type": "value",
                        "key": "CNAME",
                        "op": "regex",
                        "value": ".*inactive.*",
                    }
                ],
                "actions": [{"type": "remove-tag", "tags": ["tagTestKey"]}],
            },
            session_factory=factory,
        )
        if self.recording:
            resources = p.run()
        else:
            with mock_datetime_now(date_parse("2017-11-10"), datetime):
                resources = p.run()
        self.assertEqual(len(resources), 1)
        self.assertEqual(resources[0]["EnvironmentName"], envname)
        if self.recording:
            sleep(5)
        while self.query_env_status(factory(), envname) != "Ready":
            if self.recording:
                sleep(30)
            pass
        self.assertIsNone(self.env_tags_dict(factory(), envarn).get("tagTestKey"))



# Copyright 2016-2017 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import, division, print_function, unicode_literals

from .common import BaseTest, TestConfig as Config


class ElasticSearch(BaseTest):

    def test_resource_manager(self):
        factory = self.replay_flight_data("test_elasticsearch_query")
        p = self.load_policy(
            {
                "name": "es-query",
                "resource": "elasticsearch",
                "filters": [{"DomainName": "c7n-test"}],
            },
            session_factory=factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 1)
        self.assertEqual(resources[0]["DomainName"], "c7n-test")
        self.assertEqual(resources[0]["Tags"], [{u"Key": u"Env", u"Value": u"Dev"}])
        self.assertTrue(
            resources[0]["Endpoint"].startswith(
                "search-c7n-test-ug4l2nqtnwwrktaeagxsqso"
            )
        )

    def test_metrics_domain(self):
        factory = self.replay_flight_data("test_elasticsearch_delete")
        p = self.load_policy(
            {
                "name": "es-query",
                "resource": "elasticsearch",
                "filters": [
                    {
                        "type": "metrics",
                        "name": "SearchableDocuments",
                        "days": 4,
                        "period": 86400,
                        "value": 1000,
                        "op": "less-than",
                    }
                ],
            },
            config=Config.empty(),
            session_factory=factory,
        )
        self.assertEqual(
            p.resource_manager.filters[0].get_dimensions({"DomainName": "foo"}),
            [
                {"Name": "ClientId", "Value": "644160558196"},
                {"Name": "DomainName", "Value": "foo"},
            ],
        )

    def test_delete_search(self):
        factory = self.replay_flight_data("test_elasticsearch_delete")
        p = self.load_policy(
            {
                "name": "es-query",
                "resource": "elasticsearch",
                "filters": [{"DomainName": "c7n-test"}],
                "actions": ["delete"],
            },
            session_factory=factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 1)
        self.assertEqual(resources[0]["DomainName"], "c7n-test")

        client = factory().client("es")

        state = client.describe_elasticsearch_domain(DomainName="c7n-test")[
            "DomainStatus"
        ]
        self.assertEqual(state["Deleted"], True)

    def test_domain_add_tag(self):
        session_factory = self.replay_flight_data("test_elasticsearch_add_tag")
        client = session_factory(region="us-east-1").client("es")
        p = self.load_policy(
            {
                "name": "tag-elasticsearch-domain",
                "resource": "elasticsearch",
                "filters": [{"tag:MyTag": "absent"}],
                "actions": [{"type": "tag", "key": "MyTag", "value": "MyValue"}],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 1)
        self.assertEqual(resources[0]["DomainName"], "c7n-test")
        tags = client.list_tags(ARN=resources[0]["ARN"])["TagList"][0]
        self.assertEqual(tags, {"Key": "MyTag", "Value": "MyValue"})

    def test_domain_remove_tag(self):
        session_factory = self.replay_flight_data("test_elasticsearch_remove_tag")
        client = session_factory(region="us-east-1").client("es")
        p = self.load_policy(
            {
                "name": "remove-tag-elasticsearch-domain",
                "resource": "elasticsearch",
                "filters": [{"tag:MyTag": "present"}],
                "actions": [{"type": "remove-tag", "tags": ["MyTag"]}],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 1)
        self.assertEqual(resources[0]["DomainName"], "c7n-test")
        tags = client.list_tags(ARN=resources[0]["ARN"])["TagList"]
        self.assertEqual(len(tags), 0)

    def test_domain_mark_for_op(self):
        session_factory = self.replay_flight_data("test_elasticsearch_markforop")
        client = session_factory(region="us-east-1").client("es")
        p = self.load_policy(
            {
                "name": "markforop-elasticsearch-domain",
                "resource": "elasticsearch",
                "filters": [{"tag:MyTag": "absent"}],
                "actions": [
                    {
                        "type": "mark-for-op",
                        "days": 1,
                        "tag": "es_custodian_cleanup",
                        "op": "delete",
                    }
                ],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 1)
        self.assertEqual(resources[0]["DomainName"], "c7n-test")
        tags = client.list_tags(ARN=resources[0]["ARN"])["TagList"][0]
        self.assertEqual(
            tags,
            {
                "Key": "es_custodian_cleanup",
                "Value": "Resource does not meet policy: delete@2017/11/30",
            },
        )

    def test_domain_marked_for_op(self):
        session_factory = self.replay_flight_data("test_elasticsearch_markedforop")
        p = self.load_policy(
            {
                "name": "markedforop-elasticsearch-domain",
                "resource": "elasticsearch",
                "filters": [
                    {
                        "type": "marked-for-op",
                        "tag": "es_custodian_cleanup",
                        "skew": 1,
                        "op": "delete",
                    }
                ],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        self.assertEqual(len(resources), 1)
        self.assertEqual(resources[0]["DomainName"], "c7n-test")

    def test_modify_security_groups(self):
        session_factory = self.replay_flight_data(
            "test_elasticsearch_modify_security_groups"
        )
        p = self.load_policy(
            {
                "name": "modify-es-sg",
                "resource": "elasticsearch",
                "filters": [
                    {
                        "type": "security-group",
                        "key": "GroupId",
                        "value": ["sg-6c7fa917", "sg-3839ec4b"],
                        "op": "in",
                    }
                ],
                "actions": [
                    {
                        "type": "modify-security-groups",
                        "add": ["sg-9a5386e9"],
                        "remove": ["sg-3839ec4b"],
                    }
                ],
            },
            session_factory=session_factory,
        )
        resources = p.run()
        self.assertEqual(
            sorted(resources[0]["VPCOptions"]["SecurityGroupIds"]),
            sorted(["sg-6c7fa917", "sg-3839ec4b"]),
        )

        client = session_factory(region="us-east-1").client("es")
        result = client.describe_elasticsearch_domains(
            DomainNames=[resources[0]["DomainName"]]
        )[
            "DomainStatusList"
        ]
        self.assertEqual(
            sorted(result[0]["VPCOptions"]["SecurityGroupIds"]),
            sorted(["sg-6c7fa917", "sg-9a5386e9"]),
        )



# Copyright 2016-2017 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import, division, print_function, unicode_literals

import time
import unittest
import logging

from c7n.log import CloudWatchLogHandler
from .common import BaseTest


class LogTest(BaseTest):

    def test_existing_stream(self):
        session_factory = self.replay_flight_data("test_log_existing_stream")
        client = session_factory().client("logs")
        group_name = "/custodian-dev"
        client.create_log_group(logGroupName=group_name)
        handler = CloudWatchLogHandler(group_name, session_factory=session_factory)
        log = logging.getLogger("custodian")
        log.addHandler(handler)
        self.addCleanup(log.removeHandler, handler)
        log.setLevel(logging.DEBUG)

        for i in range(100, 115):
            log.info("hello world %s" % i)

        handler.flush()
        handler.close()

    def test_time_flush(self):
        session_factory = self.replay_flight_data("test_log_time_flush")
        log = logging.getLogger("test-c7n")
        handler = CloudWatchLogHandler(
            "test-c7n-4", "alpha", session_factory=session_factory
        )
        handler.batch_interval = 0.1
        log.addHandler(handler)
        self.addCleanup(log.removeHandler, handler)
        log.setLevel(logging.DEBUG)

        for i in range(100, 105):
            log.info("hello world %s" % i)

        time.sleep(0.2)
        log.info("bye world")
        self.assertFalse(handler.buf)
        handler.flush()
        handler.close()

    def test_transport_buffer_flush(self):
        session_factory = self.replay_flight_data("test_transport_buffer_flush")
        log = logging.getLogger("test-c7n")
        handler = CloudWatchLogHandler(
            "test-c7n-5", "alpha", session_factory=session_factory
        )
        handler.batch_size = 5
        log.addHandler(handler)
        self.addCleanup(log.removeHandler, handler)
        log.setLevel(logging.DEBUG)

        for i in range(10):
            log.info("knock, knock %d" % i)

        handler.flush()
        self.assertFalse(handler.transport.buffers)


if __name__ == "__main__":
    unittest.main()



# Copyright 2016-2017 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Query capability built on skew metamodel

tags_spec -> s3, elb, rds
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import functools
import itertools
import json
from concurrent.futures import as_completed

import jmespath
import six


from c7n.actions import ActionRegistry
from c7n.exceptions import ClientError, ResourceLimitExceeded
from c7n.filters import FilterRegistry, MetricsFilter
from c7n.manager import ResourceManager
from c7n.registry import PluginRegistry
from c7n.tags import register_ec2_tags, register_universal_tags
from c7n.utils import (
    local_session, generate_arn, get_retry, chunks, camelResource)


try:
    from botocore.paginate import PageIterator
except ImportError:
    # Likely using another provider in a serverless environment
    class PageIterator(object):
        pass


class ResourceQuery(object):

    def __init__(self, session_factory):
        self.session_factory = session_factory

    @staticmethod
    def resolve(resource_type):
        if not isinstance(resource_type, type):
            raise ValueError(resource_type)
        else:
            m = resource_type
        return m

    def _invoke_client_enum(self, client, enum_op, params, path, retry=None):
        if client.can_paginate(enum_op):
            p = client.get_paginator(enum_op)
            if retry:
                p.PAGE_ITERATOR_CLS = RetryPageIterator
            results = p.paginate(**params)
            data = results.build_full_result()
        else:
            op = getattr(client, enum_op)
            data = op(**params)

        if path:
            path = jmespath.compile(path)
            data = path.search(data)

        return data

    def filter(self, resource_manager, **params):
        """Query a set of resources."""
        m = self.resolve(resource_manager.resource_type)
        client = local_session(self.session_factory).client(
            m.service, resource_manager.config.region)
        enum_op, path, extra_args = m.enum_spec
        if extra_args:
            params.update(extra_args)
        return self._invoke_client_enum(
            client, enum_op, params, path,
            getattr(resource_manager, 'retry', None)) or []

    def get(self, resource_manager, identities):
        """Get resources by identities
        """
        m = self.resolve(resource_manager.resource_type)
        params = {}
        client_filter = False

        # Try to formulate server side query
        if m.filter_name:
            if m.filter_type == 'list':
                params[m.filter_name] = identities
            elif m.filter_type == 'scalar':
                assert len(identities) == 1, "Scalar server side filter"
                params[m.filter_name] = identities[0]
        else:
            client_filter = True

        resources = self.filter(resource_manager, **params)
        if client_filter:
            # This logic was added to prevent the issue from:
            # https://github.com/capitalone/cloud-custodian/issues/1398
            if all(map(lambda r: isinstance(r, six.string_types), resources)):
                resources = [r for r in resources if r in identities]
            else:
                resources = [r for r in resources if r[m.id] in identities]

        return resources


class ChildResourceQuery(ResourceQuery):
    """A resource query for resources that must be queried with parent information.

    Several resource types can only be queried in the context of their
    parents identifiers. ie. efs mount targets (parent efs), route53 resource
    records (parent hosted zone), ecs services (ecs cluster).
    """

    capture_parent_id = False
    parent_key = 'c7n:parent-id'

    def __init__(self, session_factory, manager):
        self.session_factory = session_factory
        self.manager = manager

    def filter(self, resource_manager, **params):
        """Query a set of resources."""
        m = self.resolve(resource_manager.resource_type)
        client = local_session(self.session_factory).client(m.service)

        enum_op, path, extra_args = m.enum_spec
        if extra_args:
            params.update(extra_args)

        parent_type, parent_key, annotate_parent = m.parent_spec
        parents = self.manager.get_resource_manager(parent_type)
        parent_ids = [p[parents.resource_type.id] for p in parents.resources()]

        # Bail out with no parent ids...
        existing_param = parent_key in params
        if not existing_param and len(parent_ids) == 0:
            return []

        # Handle a query with parent id
        if existing_param:
            return self._invoke_client_enum(client, enum_op, params, path)

        # Have to query separately for each parent's children.
        results = []
        for parent_id in parent_ids:
            merged_params = self.get_parent_parameters(params, parent_id, parent_key)
            subset = self._invoke_client_enum(
                client, enum_op, merged_params, path, retry=self.manager.retry)
            if annotate_parent:
                for r in subset:
                    r[self.parent_key] = parent_id
            if subset and self.capture_parent_id:
                results.extend([(parent_id, s) for s in subset])
            elif subset:
                results.extend(subset)
        return results

    def get_parent_parameters(self, params, parent_id, parent_key):
        return dict(params, **{parent_key: parent_id})


class QueryMeta(type):

    def __new__(cls, name, parents, attrs):
        if 'resource_type' not in attrs:
            return super(QueryMeta, cls).__new__(cls, name, parents, attrs)

        if 'filter_registry' not in attrs:
            attrs['filter_registry'] = FilterRegistry(
                '%s.filters' % name.lower())
        if 'action_registry' not in attrs:
            attrs['action_registry'] = ActionRegistry(
                '%s.actions' % name.lower())

        if attrs['resource_type']:
            m = ResourceQuery.resolve(attrs['resource_type'])
            # Generic cloud watch metrics support
            if m.dimension:
                attrs['filter_registry'].register('metrics', MetricsFilter)
            # EC2 Service boilerplate ...
            if m.service == 'ec2':
                # Generic ec2 resource tag support
                if getattr(m, 'taggable', True):
                    register_ec2_tags(
                        attrs['filter_registry'], attrs['action_registry'])
            if getattr(m, 'universal_taggable', False):
                register_universal_tags(
                    attrs['filter_registry'], attrs['action_registry'])

        return super(QueryMeta, cls).__new__(cls, name, parents, attrs)


def _napi(op_name):
    return op_name.title().replace('_', '')


sources = PluginRegistry('sources')


@sources.register('describe')
class DescribeSource(object):

    QueryFactory = ResourceQuery

    def __init__(self, manager):
        self.manager = manager
        self.query = self.QueryFactory(self.manager.session_factory)

    def get_resources(self, ids, cache=True):
        return self.query.get(self.manager, ids)

    def resources(self, query):
        return self.query.filter(self.manager, **query)

    def get_permissions(self):
        m = self.manager.get_model()
        perms = ['%s:%s' % (m.service, _napi(m.enum_spec[0]))]
        if getattr(m, 'detail_spec', None):
            perms.append("%s:%s" % (m.service, _napi(m.detail_spec[0])))
        if getattr(m, 'batch_detail_spec', None):
            perms.append("%s:%s" % (m.service, _napi(m.batch_detail_spec[0])))
        return perms

    def augment(self, resources):
        model = self.manager.get_model()
        if getattr(model, 'detail_spec', None):
            detail_spec = getattr(model, 'detail_spec', None)
            _augment = _scalar_augment
        elif getattr(model, 'batch_detail_spec', None):
            detail_spec = getattr(model, 'batch_detail_spec', None)
            _augment = _batch_augment
        else:
            return resources
        _augment = functools.partial(
            _augment, self.manager, model, detail_spec)
        with self.manager.executor_factory(
                max_workers=self.manager.max_workers) as w:
            results = list(w.map(
                _augment, chunks(resources, self.manager.chunk_size)))
            return list(itertools.chain(*results))


@sources.register('describe-child')
class ChildDescribeSource(DescribeSource):

    resource_query_factory = ChildResourceQuery

    def __init__(self, manager):
        self.manager = manager
        self.query = self.get_query()

    def get_query(self):
        return self.resource_query_factory(
            self.manager.session_factory, self.manager)


@sources.register('config')
class ConfigSource(object):

    retry = staticmethod(get_retry(('ThrottlingException',)))

    def __init__(self, manager):
        self.manager = manager

    def get_permissions(self):
        return ["config:GetResourceConfigHistory",
                "config:ListDiscoveredResources"]

    def get_resources(self, ids, cache=True):
        client = local_session(self.manager.session_factory).client('config')
        results = []
        m = self.manager.get_model()
        for i in ids:
            revisions = self.retry(
                client.get_resource_config_history,
                resourceId=i,
                resourceType=m.config_type,
                limit=1).get('configurationItems')
            if not revisions:
                continue
            results.append(self.load_resource(revisions[0]))
        return filter(None, results)

    def load_resource(self, item):
        if isinstance(item['configuration'], six.string_types):
            item_config = json.loads(item['configuration'])
        else:
            item_config = item['configuration']
        return camelResource(item_config)

    def resources(self, query=None):
        client = local_session(self.manager.session_factory).client('config')
        paginator = client.get_paginator('list_discovered_resources')
        paginator.PAGE_ITERATOR_CLS = RetryPageIterator
        pages = paginator.paginate(
            resourceType=self.manager.get_model().config_type)
        results = []

        with self.manager.executor_factory(max_workers=5) as w:
            ridents = pages.build_full_result()
            resource_ids = [
                r['resourceId'] for r in ridents.get('resourceIdentifiers', ())]
            self.manager.log.debug(
                "querying %d %s resources",
                len(resource_ids),
                self.manager.__class__.__name__.lower())

            for resource_set in chunks(resource_ids, 50):
                futures = []
                futures.append(w.submit(self.get_resources, resource_set))
                for f in as_completed(futures):
                    if f.exception():
                        self.manager.log.error(
                            "Exception getting resources from config \n %s" % (
                                f.exception()))
                    results.extend(f.result())
        return results

    def augment(self, resources):
        return resources


@six.add_metaclass(QueryMeta)
class QueryResourceManager(ResourceManager):

    resource_type = ""

    retry = None

    # TODO Check if we can move to describe source
    max_workers = 3
    chunk_size = 20

    permissions = ()

    _generate_arn = None

    retry = staticmethod(
        get_retry((
            'ThrottlingException',
            'RequestLimitExceeded',
            'Throttled',
            'Throttling',
            'Client.RequestLimitExceeded')))

    def __init__(self, data, options):
        super(QueryResourceManager, self).__init__(data, options)
        self.source = self.get_source(self.source_type)

    @property
    def source_type(self):
        return self.data.get('source', 'describe')

    def get_source(self, source_type):
        return sources.get(source_type)(self)

    @classmethod
    def get_model(cls):
        return ResourceQuery.resolve(cls.resource_type)

    @classmethod
    def match_ids(cls, ids):
        """return ids that match this resource type's id format."""
        id_prefix = getattr(cls.get_model(), 'id_prefix', None)
        if id_prefix is not None:
            return [i for i in ids if i.startswith(id_prefix)]
        return ids

    def get_permissions(self):
        perms = self.source.get_permissions()
        if getattr(self, 'permissions', None):
            perms.extend(self.permissions)
        return perms

    def get_cache_key(self, query):
        return {
            'account': self.account_id,
            'region': self.config.region,
            'resource': str(self.__class__.__name__),
            'q': query
        }

    def resources(self, query=None):
        cache_key = self.get_cache_key(query)
        resources = None

        if self._cache.load():
            resources = self._cache.get(cache_key)
            if resources is not None:
                self.log.debug("Using cached %s: %d" % (
                    "%s.%s" % (self.__class__.__module__,
                               self.__class__.__name__),
                    len(resources)))

        if resources is None:
            if query is None:
                query = {}
            with self.ctx.tracer.subsegment('resource-fetch'):
                resources = self.source.resources(query)
            with self.ctx.tracer.subsegment('resource-augment'):
                resources = self.augment(resources)
            self._cache.save(cache_key, resources)

        resource_count = len(resources)
        with self.ctx.tracer.subsegment('filter'):
            resources = self.filter_resources(resources)

        # Check if we're out of a policies execution limits.
        if self.data == self.ctx.policy.data:
            self.check_resource_limit(len(resources), resource_count)
        return resources

    def check_resource_limit(self, selection_count, population_count):
        """Check if policy's execution affects more resources then its limit.

        Ideally this would be at a higher level but we've hidden
        filtering behind the resource manager facade for default usage.
        """
        p = self.ctx.policy
        if isinstance(p.max_resources, int) and selection_count > p.max_resources:
            raise ResourceLimitExceeded(
                ("policy: %s exceeded resource limit: {limit} "
                 "found: {selection_count}") % p.name,
                "max-resources", p.max_resources, selection_count, population_count)
        elif p.max_resources_percent:
            if (population_count * (
                    p.max_resources_percent / 100.0) < selection_count):
                raise ResourceLimitExceeded(
                    ("policy: %s exceeded resource limit: {limit}%% "
                     "found: {selection_count} total: {population_count}") % p.name,
                    "max-percent", p.max_resources_percent, selection_count, population_count)
        return True

    def _get_cached_resources(self, ids):
        key = self.get_cache_key(None)
        if self._cache.load():
            resources = self._cache.get(key)
            if resources is not None:
                self.log.debug("Using cached results for get_resources")
                m = self.get_model()
                id_set = set(ids)
                return [r for r in resources if r[m.id] in id_set]
        return None

    def get_resources(self, ids, cache=True, augment=True):
        if cache:
            resources = self._get_cached_resources(ids)
            if resources is not None:
                return resources
        try:
            resources = self.source.get_resources(ids)
            if augment:
                resources = self.augment(resources)
            return resources
        except ClientError as e:
            self.log.warning("event ids not resolved: %s error:%s" % (ids, e))
            return []

    def augment(self, resources):
        """subclasses may want to augment resources with additional information.

        ie. we want tags by default (rds, elb), and policy, location, acl for
        s3 buckets.
        """
        return self.source.augment(resources)

    @property
    def account_id(self):
        """ Return the current account ID.

        This should now be passed in using the --account-id flag, but for a
        period of time we will support the old behavior of inferring this from
        IAM.
        """
        return self.config.account_id

    @property
    def region(self):
        """ Return the current region.
        """
        return self.config.region

    def get_arns(self, resources):
        arns = []

        m = self.get_model()
        arn_key = getattr(m, 'arn', None)
        id_key = m.id

        for r in resources:
            _id = r[id_key]
            if arn_key:
                arns.append(r[arn_key])
            elif 'arn' in _id[:3]:
                arns.append(_id)
            else:
                arns.append(self.generate_arn(_id))
        return arns

    @property
    def generate_arn(self):
        """ Generates generic arn if ID is not already arn format.
        """
        if self._generate_arn is None:
            self._generate_arn = functools.partial(
                generate_arn,
                self.get_model().service,
                region=self.config.region,
                account_id=self.account_id,
                resource_type=self.get_model().type,
                separator='/')
        return self._generate_arn


class ChildResourceManager(QueryResourceManager):

    child_source = 'describe-child'

    @property
    def source_type(self):
        source = self.data.get('source', self.child_source)
        if source == 'describe':
            source = self.child_source
        return source

    def get_parent_manager(self):
        return self.get_resource_manager(self.resource_type.parent_spec[0])


def _batch_augment(manager, model, detail_spec, resource_set):
    detail_op, param_name, param_key, detail_path, detail_args = detail_spec
    client = local_session(manager.session_factory).client(
        model.service, region_name=manager.config.region)
    op = getattr(client, detail_op)
    if manager.retry:
        args = (op,)
        op = manager.retry
    else:
        args = ()
    kw = {param_name: [param_key and r[param_key] or r for r in resource_set]}
    if detail_args:
        kw.update(detail_args)
    response = op(*args, **kw)
    return response[detail_path]


def _scalar_augment(manager, model, detail_spec, resource_set):
    detail_op, param_name, param_key, detail_path = detail_spec
    client = local_session(manager.session_factory).client(
        model.service, region_name=manager.config.region)
    op = getattr(client, detail_op)
    if manager.retry:
        args = (op,)
        op = manager.retry
    else:
        args = ()
    results = []
    for r in resource_set:
        kw = {param_name: param_key and r[param_key] or r}
        response = op(*args, **kw)
        if detail_path:
            response = response[detail_path]
        else:
            response.pop('ResponseMetadata')
        if param_key is None:
            response[model.id] = r
            r = response
        else:
            r.update(response)
        results.append(r)
    return results


class RetryPageIterator(PageIterator):

    retry = staticmethod(QueryResourceManager.retry)

    def _make_request(self, current_kwargs):
        return self.retry(self._method, **current_kwargs)
